from __future__ import annotations

import itertools
from dataclasses import dataclass
from functools import cmp_to_key

from rich.text import Text
from zabbix_cli.output.style import CodeBlockStyle
from zabbix_cli.output.style import CodeStyle

CODEBLOCK_STYLES = list(CodeBlockStyle)
CODE_STYLES = list(CodeStyle)
CODEBLOCK_LANGS = {
    "python": "py",
}


@dataclass
class MarkdownSpan:
    start: int
    end: int
    italic: bool = False
    bold: bool = False
    code: bool = False
    codeblock: bool = False
    language: str = ""

    def to_symbols(self) -> tuple[MarkdownSymbol, MarkdownSymbol]:
        start = MarkdownSymbol.from_span(self, end=False)
        end = MarkdownSymbol.from_span(self, end=True)
        return start, end


@dataclass
class MarkdownSymbol:
    position: int
    italic: bool = False
    bold: bool = False
    code: bool = False
    codeblock: bool = False
    end: bool = False
    language: str = ""

    @property
    def symbol(self) -> str:
        symbol: list[str] = []
        if self.codeblock:
            # Only insert language when opening codeblock
            lang = self.language if not self.end else ""
            symbol.append(f"```{lang}\n")
            # TODO: add support for language in fences (codeblock)
        else:
            if self.italic:
                symbol.append("*")
            if self.bold:
                symbol.append("**")
            if self.code:
                symbol.append("`")
        s = "".join(symbol)
        if self.end:
            s = f"{s[::-1]}"
        return s

    @classmethod
    def from_span(cls, span: MarkdownSpan, *, end: bool = False) -> MarkdownSymbol:
        return cls(
            position=span.end if end else span.start,
            italic=span.italic,
            bold=span.bold,
            code=span.code,
            codeblock=span.codeblock,
            end=end,
            language=span.language,
        )


# Easier than implementing rich comparison methods on MarkdownSymbol
def mdsymbol_cmp(a: MarkdownSymbol, b: MarkdownSymbol) -> int:
    if a.position < b.position:
        return -1
    elif a.position > b.position:
        return 1
    else:
        # code tags cannot have other tags inside them
        if a.code and not b.code:
            return 1
        if b.code and not a.code:
            return -1
    return 0


# TODO: rename `markup_to_markdown` to `markup_as_markdown`
# OR    rename `markup_to_plaintext` to `markup_as_plaintext`
#       I am partial to `x_to_y`.


def markup_to_markdown(s: str) -> str:
    """Parses a string that might contain markup formatting and converts it to Markdown.

    This is a very naive implementation that only supports a subset of Rich markup, but it's
    good enough for our purposes.
    """
    t = Text.from_markup(normalize_spaces(s))
    spans: list[MarkdownSpan] = []
    # Markdown has more limited styles than Rich markup, so we just
    # identify the ones we care about and ignore the rest.
    for span in t.spans:
        new_span = MarkdownSpan(span.start, span.end)
        styles = str(span.style).lower().split(" ")
        # Code (block) styles ignore other styles
        if any(s in CODEBLOCK_STYLES for s in styles):
            new_span.codeblock = True
            lang = next((s for s in styles if s in CODEBLOCK_LANGS), "")
            new_span.language = CODEBLOCK_LANGS.get(lang, "")
        elif any(s in CODE_STYLES for s in styles):
            new_span.code = True
        else:
            if "italic" in styles:
                new_span.italic = True
            if "bold" in styles:
                new_span.bold = True
        spans.append(new_span)

    # Convert MarkdownSpans to MarkdownSymbols
    # Each MarkdownSymbol represents a markdown formatting character along
    # with its position in the string.
    symbols = list(itertools.chain.from_iterable(sp.to_symbols() for sp in spans))
    symbols = sorted(symbols, key=cmp_to_key(mdsymbol_cmp))

    # List of characters that make up string
    plaintext = list(str(t.plain.strip()))  # remove leading and trailing whitespace
    offset = 0
    for symbol in symbols:
        plaintext.insert(symbol.position + offset, symbol.symbol)
        offset += 1

    return "".join(plaintext)


def normalize_spaces(s: str) -> str:
    """Normalizes spaces in a string while keeping newlines intact."""
    split = filter(None, s.split(" "))
    parts: list[str] = []
    for part in split:
        if part.endswith("\n"):
            parts.append(part)
        else:
            parts.append(f"{part} ")
    return "".join(parts)


def markup_as_plain_text(s: str) -> str:
    """Renders a string that might contain markup formatting as a plain text string."""
    return Text.from_markup(s).plain



from __future__ import annotations

from typing import Optional

import typer

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.output.console import err_console
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import TriggerPriority
from zabbix_cli.utils.args import parse_bool_arg
from zabbix_cli.utils.args import parse_int_arg
from zabbix_cli.utils.args import parse_list_arg

HELP_PANEL = "Problem"


@app.command(name="acknowledge_event", rich_help_panel=HELP_PANEL)
def acknowledge_event(
    ctx: typer.Context,
    event_ids: str = typer.Argument(
        help="Comma-separated list of event ID(s)",
        show_default=False,
    ),
    message: str = typer.Option(
        "[Zabbix-CLI] Acknowledged via acknowledge_events",
        "--message",
        "-m",
        help="Message to add to the event",
    ),
    close: bool = typer.Option(
        False,
        "--close",
        "-c",
        help="Close the event after acknowledging it",
    ),
    # Legacy positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Acknowledge events by ID."""
    from zabbix_cli.commands.results.problem import AcknowledgeEventResult
    from zabbix_cli.models import Result

    eids = parse_list_arg(event_ids)
    if not eids:
        exit_err("No event IDs specified.")

    if args:
        if len(args) != 2:
            exit_err("Invalid number of positional arguments.")
        message = args[0]
        close = parse_bool_arg(args[1])
    # Don't prompt for missing message. It's optional.
    acknowledged_ids = app.state.client.acknowledge_event(
        *eids, message=message, close=close
    )

    msg = "Event(s) acknowledged"
    if close:
        msg += " and closed"
    render_result(
        Result(
            message=msg,
            result=AcknowledgeEventResult(
                event_ids=acknowledged_ids, close=close, message=message
            ),
        )
    )


@app.command(
    name="acknowledge_trigger_last_event",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Acknowledge the last event for trigger 12345",
            "acknowledge_trigger_last_event 12345",
        ),
        Example(
            "Acknowledge the last event for trigger 12345 with a message and close it",
            "acknowledge_trigger_last_event 12345 --message 'Acked via CLI' --close",
        ),
        Example(
            "Acknowledge multiple triggers",
            "acknowledge_trigger_last_event 12345,12346",
        ),
    ],
)
def acknowledge_trigger_last_event(
    ctx: typer.Context,
    trigger_ids: str,
    message: Optional[str] = typer.Option(
        None,
        "--message",
        help="Acknowledgement message",
    ),
    close: bool = typer.Option(
        False,
        "--close",
        "-c",
        help="Close event",
    ),
    # Legacy positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Acknowledge the the last event for the given triggers."""
    from zabbix_cli.commands.results.problem import AcknowledgeTriggerLastEventResult
    from zabbix_cli.models import Result

    tids = parse_list_arg(trigger_ids)
    if not tids:
        exit_err("No trigger IDs specified.")
    if args:
        if len(args) != 2:
            exit_err("Invalid number of positional arguments.")
        message = args[0]
        close = parse_bool_arg(args[1])

    # Message is optional, so we don't prompt for it if it's missing
    events = [app.state.client.get_event(object_id=tid) for tid in tids]
    event_ids = [e.eventid for e in events]
    acknowledged_ids = app.state.client.acknowledge_event(
        *event_ids, message=message, close=close
    )

    msg = "Event(s) acknowledged"
    if close:
        msg += " and closed"
    render_result(
        Result(
            message=msg,
            result=AcknowledgeTriggerLastEventResult(
                trigger_ids=tids,
                event_ids=acknowledged_ids,
                close=close,
                message=message,
            ),
        )
    )


@app.command(name="show_alarms", rich_help_panel=HELP_PANEL)
def show_alarms(
    ctx: typer.Context,
    description: Optional[str] = typer.Option(
        None,
        "--description",
        help="Description of alarm(s) to show.",
    ),
    # Could this be a list of priorities in V2?
    priority: Optional[TriggerPriority] = typer.Option(
        None,
        "--priority",
        help="Priority of alarm(s) to show.",
    ),
    hostgroups: Optional[str] = typer.Option(
        None,
        "--hostgroup",
        help="Host group(s) to show alarms for. Comma-separated.",
    ),
    unacknowledged: bool = typer.Option(
        True,
        "--unack/--ack",
        help="Show only alarms whose last event is unacknowledged.",
    ),
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Show the latest events for the given triggers, hosts, and/or host groups.

    At least one trigger ID, host or host group must be specified.
    """
    from zabbix_cli.models import AggregateResult

    if args:
        if len(args) != 4:
            exit_err("Invalid number of positional arguments.")
        description = args[0]
        priority = TriggerPriority(args[1])
        hostgroups = args[2]
        # in V2, "*" was used to indicate "false"
        if args[3] == "*":
            unacknowledged = False
        else:
            unacknowledged = parse_bool_arg(args[3])

    hostgroups_args = parse_list_arg(hostgroups)
    hgs = [app.state.client.get_hostgroup(hg) for hg in hostgroups_args]
    with app.status("Fetching triggers..."):
        triggers = app.state.client.get_triggers(
            hostgroups=hgs,
            description=description,
            priority=priority,
            unacknowledged=unacknowledged,
            select_hosts=True,
            skip_dependent=True,
            monitored=True,
            active=True,
            expand_description=True,
            filter={"value": 1},  # why?
        )
    render_result(AggregateResult(result=triggers))


@app.command(
    name="show_trigger_events",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show recent events for host foo.example.com",
            "show_trigger_events --host foo.example.com",
        ),
        Example(
            "Show recent events for hosts in host group 'Linux servers'",
            "show_trigger_events --hostgroup 'Linux servers'",
        ),
        Example(
            "Show 20 most recent events for triggers 12345 & 12346",
            "show_trigger_events --trigger-id 12345,12346 --limit 20",
        ),
    ],
)
def show_trigger_events(
    ctx: typer.Context,
    trigger_id: Optional[str] = typer.Option(
        None,
        "--trigger-id",
        help="ID of trigger(s) to show events for.",
    ),
    hostgroups: Optional[str] = typer.Option(
        None,
        "--hostgroup",
        help="Host group(s) to show events for.",
    ),
    hosts: Optional[str] = typer.Option(
        None,
        "--host",
        help="Host(s) to show events for.",
    ),
    limit: int = typer.Option(
        10,
        "--limit",
        "-l",
        help="Maximum number of events to show.",
    ),
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Show the latest events for the given triggers, hosts, and/or host groups.

    At least one trigger ID, host or host group must be specified.
    """
    from zabbix_cli.models import AggregateResult

    if args:
        if len(args) != 2:
            exit_err("Invalid number of positional arguments.")
        trigger_id = args[0]
        limit = parse_int_arg(args[1])

    # Parse commma-separated args
    trigger_ids = parse_list_arg(trigger_id)
    hostgroups_args = parse_list_arg(hostgroups)
    hosts_args = parse_list_arg(hosts)
    if not trigger_ids and not hostgroups_args and not hosts_args:
        err_console.print(ctx.get_help())
        exit_err("At least one trigger ID, host or host group must be specified.")

    # Fetch the host(group)s if specified
    hostgroups_list = [app.state.client.get_hostgroup(hg) for hg in hostgroups_args]
    hosts_list = [app.state.client.get_host(host) for host in hosts_args]

    with app.status("Fetching events..."):
        events = app.state.client.get_events(
            object_ids=trigger_ids,
            group_ids=[hg.groupid for hg in hostgroups_list],
            host_ids=[host.hostid for host in hosts_list],
            sort_field="clock",
            sort_order="DESC",
            limit=limit,
        )
    render_result(AggregateResult(result=events))



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

from packaging.version import Version
from pydantic import ConfigDict
from pydantic import FieldSerializationInfo
from pydantic import field_serializer
from rich.box import SIMPLE_HEAD
from rich.table import Table
from typing_extensions import Self
from typing_extensions import TypedDict

from zabbix_cli.models import TableRenderable
from zabbix_cli.output.console import error
from zabbix_cli.output.formatting.path import path_link

if TYPE_CHECKING:
    from zabbix_cli.commands.cli import DirectoryType
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType
    from zabbix_cli.state import State


class ImplementationInfo(TypedDict):
    name: str
    version: tuple[Any, ...]
    hexversion: int
    cache_tag: str


class PythonInfo(TypedDict):
    version: str
    implementation: ImplementationInfo
    platform: str


class DebugInfo(TableRenderable):
    config_path: Optional[Path] = None
    api_version: Optional[Version] = None
    url: Optional[str] = None
    user: Optional[str] = None
    auth_token: Optional[str] = None
    connected_to_zabbix: bool = False
    python: PythonInfo

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("api_version")
    def ser_api_version(self, _info: FieldSerializationInfo) -> str:
        return str(self.api_version)

    @property
    def config_path_str(self) -> str:
        return (
            path_link(self.config_path) if self.config_path else str(self.config_path)
        )

    @classmethod
    def from_debug_data(cls, state: State, *, with_auth: bool = False) -> DebugInfo:
        # So far we only use state, but we can expand this in the future
        from zabbix_cli.exceptions import ZabbixCLIError

        obj = cls(
            python={
                "version": sys.version,
                "implementation": {
                    "name": sys.implementation.name,
                    "version": sys.implementation.version,
                    "hexversion": sys.implementation.hexversion,
                    "cache_tag": sys.implementation.cache_tag,
                },
                "platform": sys.platform,
            }
        )

        # Config might not be configured
        try:
            obj.config_path = state.config.config_path
        except ZabbixCLIError:
            pass

        # We might not be connected to the API
        try:
            obj.api_version = state.client.version
            obj.url = state.client.url
            obj.user = state.config.api.username
            if with_auth:
                obj.auth_token = state.client.auth
            obj.connected_to_zabbix = True
        except ZabbixCLIError:
            error("Unable to retrieve API info: Not connected to Zabbix API. ")
        except Exception as e:
            error(f"Unable to retrieve API info: {e}")
        return obj

    def as_table(self) -> Table:
        table = Table(title="Debug Info")
        table.add_column("Key", justify="right", style="cyan")
        table.add_column("Value", justify="left", style="magenta")

        table.add_row("Config File", self.config_path_str)
        table.add_row("API URL", str(self.url))
        table.add_row("API Version", str(self.api_version))
        table.add_row("User", str(self.user))
        table.add_row("Auth Token", str(self.auth_token))
        table.add_row("Connected to Zabbix", str(self.connected_to_zabbix))
        table.add_row("Python Version", str(self.python["version"]))
        table.add_row("Platform", str(self.python["platform"]))

        return table


class HistoryResult(TableRenderable):
    """Result type for `show_history` command."""

    __show_lines__ = False
    __box__ = SIMPLE_HEAD

    commands: list[str] = []


class DirectoriesResult(TableRenderable):
    """Result type for `show_dirs` command."""

    directories: list[dict[str, Path]] = []

    @classmethod
    def from_directory_types(cls, dirs: list[DirectoryType]) -> Self:
        return cls(directories=[{str(d.value): d.as_path()} for d in dirs])

    def __cols_rows__(self) -> ColsRowsType:
        from zabbix_cli.output.style import Emoji

        cols = ["Type", "Path", "Exists"]
        rows: RowsType = []
        for d in self.directories:
            for key in d:
                p = d[key]
                exists = Emoji.fmt_bool(p.exists())
                rows.append([key, str(d[key]), exists])
        return cols, rows



"""Filesystem utilities."""

from __future__ import annotations

import logging
import os
import re
import subprocess
import sys
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.exceptions import ZabbixCLIFileError
from zabbix_cli.exceptions import ZabbixCLIFileNotFoundError
from zabbix_cli.output.console import print_path
from zabbix_cli.output.console import success

logger = logging.getLogger(__name__)


def read_file(file: Path) -> str:
    """Attempts to read the contents of a file."""
    if not file.exists():
        raise ZabbixCLIFileNotFoundError(f"File {file} does not exist")
    if not file.is_file():
        raise ZabbixCLIFileError(f"{file} is not a file")
    try:
        return file.read_text()
    except OSError as e:
        raise ZabbixCLIFileError(f"Unable to read file {file}") from e


def open_directory(
    directory: Path, command: Optional[str] = None, *, force: bool = False
) -> None:
    """Open directory in file explorer.

    Prints the path to the directory to stderr if no window server is detected.
    The path must be a directory, otherwise a ZabbixCLIError is raised.

    Args:
        directory (Path): The directory to open.
        command (str, optional): The command to use to open the directory. If `None`, the command is determined based on the platform.
        force (bool, optional): If `True`, open the directory even if no window server is detected. Defaults to `False`.
    """
    try:
        if not directory.exists():
            raise FileNotFoundError
        directory = directory.resolve(strict=True)
    except FileNotFoundError as e:
        raise ZabbixCLIError(f"Directory {directory} does not exist") from e
    except OSError as e:
        raise ZabbixCLIError(f"Unable to resolve symlinks for {directory}") from e
    if not directory.is_dir():
        raise ZabbixCLIError(f"{directory} is not a directory")

    spath = str(directory)
    if sys.platform == "win32":
        subprocess.run([command or "explorer", spath])
    elif sys.platform == "darwin":
        subprocess.run([command or "open", spath])
    else:  # Linux and Unix
        if not os.environ.get("DISPLAY") and not force:
            print_path(directory)
            return
        subprocess.run([command or "xdg-open", spath])
    success(f"Opened {directory}")


def mkdir_if_not_exists(path: Path) -> None:
    """Create a directory for a given path if it does not exist.

    Returns the path if it was created, otherwise None.
    """
    if path.exists():
        return
    try:
        path.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        raise ZabbixCLIFileError(f"Failed to create directory {path}: {e}") from e
    else:
        logger.info("Created directory: %s", path)


def sanitize_filename(filename: str) -> str:
    """Make a filename safe(r) for use in filesystems.

    Very naive implementation that removes illegal characters.
    Does not check for reserved names or path length.
    """
    return re.sub(r"[^\w\-.]", "_", filename)


def make_executable(path: Path) -> None:
    """Make a file executable."""
    if sys.platform == "win32":
        logger.debug("Skipping making file %s executable on Windows", path)
        return

    if not path.exists():
        raise ZabbixCLIFileNotFoundError(
            f"File {path} does not exist. Unable to make it executable."
        )
    mode = path.stat().st_mode
    new_mode = mode | (mode & 0o444) >> 2  # copy R bits to X
    if new_mode != mode:
        path.chmod(new_mode)
        logger.info("Changed file mode of %s from %o to %o", path, mode, new_mode)
    else:
        logger.debug("File %s is already executable", path)


def move_file(src: Path, dest: Path, *, mkdir: bool = True) -> None:
    """Move a file to a new location."""
    try:
        if mkdir:
            mkdir_if_not_exists(dest.parent)
        src.rename(dest)
    except Exception as e:
        raise ZabbixCLIError(f"Failed to move {src} to {dest}: {e}") from e
    else:
        logger.info("Moved %s to %s", src, dest)


@contextmanager
def temp_directory() -> Generator[Path, None, None]:
    """Context manager for creating a temporary directory.

    Ripped from: https://github.com/pypa/hatch/blob/35f8ffdacc937bdcf3b250e0be1bbdf5cde30c4c/src/hatch/utils/fs.py#L112-L117
    """
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as d:
        yield Path(d).resolve()



from __future__ import annotations

import click
import pytest
import typer
from inline_snapshot import snapshot
from zabbix_cli.app.app import StatefulApp
from zabbix_cli.commands.common.args import CommandParam


def test_command_param(ctx: typer.Context) -> None:
    param = CommandParam()
    assert param.name == "command"

    # No ctx
    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        param.convert("some-non-existent-command", None, None)
    assert exc_info.exconly() == snapshot("click.exceptions.BadParameter: No context.")

    # No value (empty string)
    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        param.convert("", None, ctx)
    assert exc_info.exconly() == snapshot(
        "click.exceptions.BadParameter: Missing command."
    )

    # No value (None)
    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        param.convert(None, None, ctx)  # type: ignore
    assert exc_info.exconly() == snapshot(
        "click.exceptions.BadParameter: Missing command."
    )

    # Command not found
    with pytest.raises(click.exceptions.BadParameter) as exc_info:
        param.convert("some-non-existent-command", None, None)
    assert exc_info.exconly() == snapshot("click.exceptions.BadParameter: No context.")


def test_command_param_in_command(
    app: StatefulApp, capsys: pytest.CaptureFixture[str]
) -> None:
    @app.command(name="help-command")
    def help_command(  # type: ignore
        ctx: typer.Context,
        cmd_arg: click.Command = typer.Argument(
            ..., help="The command to get help for."
        ),
    ) -> str:
        return cmd_arg.get_help(ctx)

    @app.command(name="other-command", help="Help for the other command.")
    def other_command(ctx: typer.Context) -> None:  # type: ignore
        pass

    cmd = typer.main.get_command(app)

    with cmd.make_context(None, ["help-command", "other-command"]) as new_ctx:
        new_ctx.info_name = "other-command"
        cmd.invoke(new_ctx)
        captured = capsys.readouterr()
        assert captured.err == snapshot("")
        # We cannot test the output with snapshot testing, because
        # the trim-trailing-whitespace pre-commit hook modifies the
        # snapshot output. Not ideal, so we have to test the relevant lines instead.
        #
        # Also, terminal styling is broken when testing outside of a terminal.
        # Thus, this minimal test.
        assert "other-command help-command [OPTIONS]" in captured.out



from __future__ import annotations

import logging
import sys
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Protocol
from typing import cast
from typing import runtime_checkable

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint
else:
    from importlib.metadata import EntryPoint


from zabbix_cli.exceptions import PluginLoadError
from zabbix_cli.exceptions import PluginPostImportError
from zabbix_cli.output.console import error

if TYPE_CHECKING:
    from zabbix_cli.config.model import Config
    from zabbix_cli.config.model import PluginConfig


logger = logging.getLogger(__name__)


@runtime_checkable
class PluginModule(Protocol):
    def __configure__(self, config: PluginConfig) -> None: ...


class PluginLoader:
    def __init__(self) -> None:
        self.plugins: dict[str, ModuleType] = {}

    def load(self, config: Config) -> None:
        self._load_plugins(config)
        self._load_plugins_from_metadata(config)

    def _load_plugins(self, config: Config) -> None:
        """Load plugins from local Python modules."""
        logger.debug("Loading plugins from modules in configuration file.")
        for name, plugin_config in config.plugins.root.items():
            if not plugin_config.module:
                logger.debug("Plugin %s has no module defined. Skipping", name)
                continue
            if not plugin_config.enabled:
                logger.debug("Plugin %s is disabled, skipping", name)
                continue

            logger.debug("Loading plugin: %s", name)
            try:
                module = load_plugin_module(name, plugin_config)
            except Exception as e:
                # If the exception is already a PluginLoadError
                # or a subclass, use that, otherwise create a new
                # PluginLoadError so we get the correct message format.
                if not isinstance(e, PluginLoadError):
                    exc = PluginLoadError(name, plugin_config)
                else:
                    exc = e
                if not plugin_config.optional:
                    raise exc from e
                else:
                    # Use message created by the PluginLoadError
                    msg = exc.args[0] if exc.args else str(exc)
                    error(msg, exc_info=True)
            else:
                self._add_plugin(name, module)

    def _load_plugins_from_metadata(self, config: Config) -> None:
        """Load plugins from Python package entry points."""
        # Use backport for Python < 3.10
        if sys.version_info < (3, 10):
            from importlib_metadata import (
                entry_points,  # pyright: ignore[reportUnknownVariableType]
            )
        else:
            from importlib.metadata import entry_points

        discovered_plugins = entry_points(group="zabbix-cli.plugins")

        # HACK: Cast Tuple[Any, ...] result to concrete type.
        # In order to pass type checking on all Python versions,
        # we need to pretend that we are developing on 3.9 and
        # using the backport. The problem is that the backport
        # does not define a generic tuple type for the result,
        # and instead just subclasses tuple. So we need to cast
        # the result to the correct type.
        # This is one of the drawbacks of running in 3.9 mode, but
        # it's necessary to ensure we don't introduce features that
        # do not exist in our minimum supported version.
        discovered_plugins = cast(tuple[EntryPoint], discovered_plugins)
        for plugin in discovered_plugins:
            conf = config.plugins.get(plugin.name)
            try:
                module = load_plugin_from_entry_point(plugin)
            except Exception as e:
                # By default, broken plugings will not break the application
                # unless they are explicitly marked as required in the config.
                # This is a deviation from the standard behavior of plugins, but
                # since these are third party plugins, we want to be more lenient
                # and not break the entire application if a plugin is broken.
                if not conf or not conf.optional:
                    raise PluginLoadError(plugin.name, conf) from e
                else:
                    error(f"Error loading plugin {plugin.name}: {e}", exc_info=True)
            else:
                self._add_plugin(plugin.name, module)

    def _add_plugin(self, name: str, module: ModuleType) -> None:
        self.plugins[name] = module
        logger.info("Plugin loaded: %s", name)

    def configure_plugins(self, config: Config) -> None:
        for name, module in self.plugins.items():
            if not isinstance(module, PluginModule):
                logger.debug("Plugin %s has no __configure__ function. Skipping", name)
                continue
            plugin_config = config.plugins.root.get(name)
            if not plugin_config:
                logger.warning(
                    "No configuration found for plugin '%s'. Cannot run __configure__",
                    name,
                )
                continue
            try:
                module.__configure__(plugin_config)
            except Exception as e:
                raise PluginPostImportError(name, plugin_config) from e


def load_plugin_module(plugin_name: str, plugin_config: PluginConfig) -> ModuleType:
    """Load plugin from a Python module."""
    name_or_path = plugin_config.module
    p = Path(name_or_path)
    if p.exists():
        logger.debug("Loading plugin %s from file: %s", plugin_name, p)
        mod = _load_module_from_file(name_or_path, p)
    else:
        logger.debug("Importing plugin %s as module '%s'", plugin_name, name_or_path)
        mod = _load_module(name_or_path)
    return mod


def _load_module_from_file(mod: str, file: Path) -> ModuleType:
    """Load a module from a file."""
    import importlib.util

    spec = importlib.util.spec_from_file_location(mod, str(file.resolve()))
    if not spec or not spec.loader:
        raise ImportError(f"Could not load module from file: {file}")
    module = importlib.util.module_from_spec(spec)
    sys.modules[mod] = module
    spec.loader.exec_module(module)
    logger.info("Loaded module from file: %s", file)
    return module


def _load_module(mod: str) -> ModuleType:
    """Load a module."""
    import importlib

    module = importlib.import_module(mod)
    logger.info("Loaded module: %s", mod)
    return module


def load_plugin_from_entry_point(entry_point: EntryPoint) -> ModuleType:
    """Load a plugin from an entry point."""
    logger.debug("Loading plugin %s from metadata", entry_point.name)
    module = entry_point.load()
    if not isinstance(module, ModuleType):
        return sys.modules[entry_point.module]
    return module



"""Uncategorized utility functions.

Some stemming from Zabbix-cli v2, while others relate to converting
values and flags from the Zabbix API.
"""

from __future__ import annotations

import re
from collections.abc import Iterable
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Final
from typing import NamedTuple
from typing import Optional
from typing import Union

from zabbix_cli.exceptions import ZabbixCLIError


# NOTE: consider setting with_code to False by default...
# The only downside is possibly breaking backwards compatibility
def _format_code(
    code: Union[str, int, None], status_map: dict[Any, str], *, with_code: bool = True
) -> str:
    status = status_map.get(code, "Unknown")
    if with_code and code is not None:
        status += f" ({code})"
    return status


# LEGACY: Kept for backwards compatibility in JSON output
# zabbix_cli.pyzabbix.enums.MaintenanceStatus has different string values
# than the ones used in Zabbix-cli v2. This function is used to map the
# the API values to the old string values when serializing to JSON.
# Should be removed when we drop support for legacy JSON output.
def get_maintenance_status(code: Optional[str], *, with_code: bool = False) -> str:
    """Get maintenance status from code."""
    maintenance_status = {"0": "No maintenance", "1": "In progress"}
    return _format_code(code, maintenance_status, with_code=with_code)


# LEGACY: Kept for backwards compatibility in JSON output
def get_monitoring_status(code: Optional[str], *, with_code: bool = False) -> str:
    """Get monitoring status from code."""
    monitoring_status = {"0": "Monitored", "1": "Not monitored"}
    return _format_code(code, monitoring_status, with_code=with_code)


def get_maintenance_active_days(schedule: int | None) -> list[str]:
    """Get maintenance day of week from code.

    The schedule bitmask is a 7-bit integer where each bit represents a day of the week.

    Args:
        schedule (int): Integer bitmask representing active days.

    Returns:
        list[str]: List of active days
    """
    if schedule is None:
        return []
    days = {
        0b0000001: "Monday",
        0b0000010: "Tuesday",
        0b0000100: "Wednesday",
        0b0001000: "Thursday",
        0b0010000: "Friday",
        0b0100000: "Saturday",
        0b1000000: "Sunday",
    }
    # Bitwise AND schedule with each DoW's bit mask
    # If the result is non-zero, the DoW is active
    active_days: list[str] = []
    for n, dow in days.items():
        if schedule & n:
            active_days.append(dow)
    return active_days


def get_maintenance_active_months(schedule: int | None) -> list[str]:
    """Get active maintenance month(s) from integer bitmask.

    The schedule bitmask is a 12-bit integer where each bit represents a month.

    Args:
        schedule (int): Integer bitmask representing active months.

    Returns:
        list[str]: List of active months.
    """
    if schedule is None:
        return []
    months = {
        0b000000000001: "January",
        0b000000000010: "February",
        0b000000000100: "March",
        0b000000001000: "April",
        0b000000010000: "May",
        0b000000100000: "June",
        0b000001000000: "July",
        0b000010000000: "August",
        0b000100000000: "September",
        0b001000000000: "October",
        0b010000000000: "November",
        0b100000000000: "December",
    }
    # Bitwise AND schedule with each month's bit mask
    # If the result is non-zero, the month is active
    active_months: list[str] = []
    for n, month in months.items():
        if schedule & n:
            active_months.append(month)
    return active_months


# NOTE: we could turn these into str Enums or Literals,
# so that it's easier to type check the values
ACKNOWLEDGE_ACTION_BITMASK: Final[dict[str, int]] = {
    "close": 0b000000001,
    "acknowledge": 0b000000010,
    "message": 0b000000100,
    "change_severity": 0b000001000,
    "unacknowledge": 0b000010000,
    "suppress": 0b000100000,
    "unsuppress": 0b001000000,
    "change_to_cause": 0b010000000,
    "change_to_symptom": 0b100000000,
}


def get_acknowledge_action_value(
    *,
    close: bool = False,
    acknowledge: bool = False,
    message: bool = False,
    change_severity: bool = False,
    unacknowledge: bool = False,
    suppress: bool = False,
    unsuppress: bool = False,
    change_to_cause: bool = False,
    change_to_symptom: bool = False,
) -> int:
    """Get acknowledge action bitmask value."""
    value = 0
    if close:
        value += ACKNOWLEDGE_ACTION_BITMASK["close"]
    if acknowledge:
        value += ACKNOWLEDGE_ACTION_BITMASK["acknowledge"]
    if message:
        value += ACKNOWLEDGE_ACTION_BITMASK["message"]
    if change_severity:
        value += ACKNOWLEDGE_ACTION_BITMASK["change_severity"]
    if unacknowledge:
        value += ACKNOWLEDGE_ACTION_BITMASK["unacknowledge"]
    if suppress:
        value += ACKNOWLEDGE_ACTION_BITMASK["suppress"]
    if unsuppress:
        value += ACKNOWLEDGE_ACTION_BITMASK["unsuppress"]
    if change_to_cause:
        value += ACKNOWLEDGE_ACTION_BITMASK["change_to_cause"]
    if change_to_symptom:
        value += ACKNOWLEDGE_ACTION_BITMASK["change_to_symptom"]
    return value


def get_acknowledge_actions(code: int) -> list[str]:
    """Get acknowledge actions from code.

    See: https://www.zabbix.com/documentation/current/en/manual/api/reference/event/acknowledge (action parameter)
    """
    # Create reverse lookup for action bitmask
    acknowledge_actions = {v: k for k, v in ACKNOWLEDGE_ACTION_BITMASK.items()}
    active_action: list[str] = []
    for n, action in acknowledge_actions.items():
        if code & n:
            active_action.append(action)
    return active_action


def compile_pattern(pattern: str) -> re.Pattern[str]:
    """Compile regex pattern."""
    try:
        p = re.compile(pattern)
    except re.error as e:
        raise ZabbixCLIError(f"Invalid regex pattern: {pattern}") from e
    return p


class TimeUnit(NamedTuple):
    """Time value."""

    unit: str
    tokens: Iterable[str]
    value: int
    """The value of the time unit in seconds."""


# NOTE: PLURAL TOKEN MUST BE LISTED FIRST
TIME_VALUE_DAY = TimeUnit("D", ["days", "day"], value=60 * 60 * 24)
TIME_VALUE_HOUR = TimeUnit("H", ["hours", "hour"], value=60 * 60)
TIME_VALUE_MINUTE = TimeUnit("M", ["minutes", "minute"], value=60)
TIME_VALUE_SECOND = TimeUnit("S", ["seconds", "second"], value=1)
TIME_VALUES = [
    TIME_VALUE_DAY,
    TIME_VALUE_HOUR,
    TIME_VALUE_MINUTE,
    TIME_VALUE_SECOND,
]


def convert_time_to_interval(time: str) -> tuple[datetime, datetime]:
    """Convert time to an interval of datetimes.

    `time` is a string that specifies a duration of time in
    one of the following formats:

    - `1d1h30m30s`
    - `1 day 1 hour 30 minutes 30 seconds`

    Any combination of the above is also valid, e.g.:

    - `1d1h30m`
    - `2 days 30 minutes`
    - `1 hour`

    The `time` string can also be a timestamp interval in the following format:

    - `2016-11-21T22:00 to 2016-11-21T23:00`

    """
    # Use a very simple heuristic to to determine if we have an interval:
    if " to " in time:
        return convert_timestamp_interval(time)
    # Fall back on parsing duration beginning from now:
    duration = convert_duration(time)
    start = datetime.now()
    end = start + duration
    return start, end


def convert_timestamp(ts: str) -> datetime:
    """Convert a datetime or timestamp string to a datetime object.

    Args:
        ts (str): Timestamp string

    Raises:
        ZabbixCLIError: String cannot be converted to a datetime object.

    Returns:
        datetime: Converted datetime object
    """
    try:
        return datetime.fromisoformat(ts)
    except ValueError:
        pass

    formats = [
        "%Y-%m-%dT%H:%M",  # Legacy format (no seconds)
        "%Y-%m-%dT%H:%M:%S",  # with T separator
        "%Y-%m-%d %H:%M:%S",  # without T separator
    ]
    for fmt in formats:
        try:
            return datetime.strptime(ts, fmt)
        except ValueError:
            pass
    raise ZabbixCLIError(f"Invalid timestamp: {ts}")


def convert_timestamp_interval(time: str) -> tuple[datetime, datetime]:
    """Convert timestamp interval to seconds.

    `time` is a string that specifies a timestamp interval in
    the following format:

    - `2016-11-21T22:00 to 2016-11-21T23:00`

    """
    start, sep, end = time.partition(" to ")
    if not sep:
        raise ZabbixCLIError(f"Invalid timestamp interval: {time}")
    return convert_timestamp(start), convert_timestamp(end)


def convert_duration(time: str) -> timedelta:
    """Convert duration to timedelta.

    `time` is a string that specifies a duration of time in
    one of the following formats:

    - `1d1h30m30s`
    - `1 day 1 hour 30 minutes 30 seconds`

    Any combination of the above is also valid, e.g.:

    - `1d1h30m`
    - `2 days 30 minutes`
    - `1 hour`
    """

    def try_convert_int(s: str) -> int:
        if not s:
            return 0
        try:
            return int(s)
        except ValueError:
            raise ZabbixCLIError(f"Invalid time value: {s}") from None

    time = time.replace(" ", "")
    for time_value in TIME_VALUES:
        # First replace full words (days, hours, minutes, seconds)
        for token in time_value.tokens:
            time = time.replace(token, time_value.unit)
        # Then replace abbreviations (d, h, m, s) with uppercase
    time = time.upper()

    # NOTE: this is very inelegant. The swapping of variables when
    # partitioning is particularly ugly.
    days, sep, rest = time.partition(TIME_VALUE_DAY.unit)
    if not sep:
        days, rest = rest, days
    hours, sep, rest = rest.partition(TIME_VALUE_HOUR.unit)
    if not sep:
        hours, rest = rest, hours
    minutes, sep, rest = rest.partition(TIME_VALUE_MINUTE.unit)
    if not sep:
        minutes, rest = rest, minutes
    seconds, sep, rest = rest.partition(TIME_VALUE_SECOND.unit)
    if rest:
        raise ZabbixCLIError(f"Invalid time value: {time}")

    td = timedelta(
        days=try_convert_int(days),
        hours=try_convert_int(hours),
        minutes=try_convert_int(minutes),
        seconds=try_convert_int(seconds),
    )
    return td


def convert_seconds_to_duration(seconds: int) -> str:
    """Convert seconds to duration string."""
    days, seconds = divmod(seconds, 60 * 60 * 24)
    hours, seconds = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(seconds, 60)

    duration = ""
    if days:
        duration += f"{days}d"
    if hours:
        duration += f"{hours}h"
    if minutes:
        duration += f"{minutes}m"
    if seconds:
        duration += f"{seconds}s"

    return duration



from __future__ import annotations

import typer

from zabbix_cli.app import app
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import MonitoringStatus

HELP_PANEL = "Host Monitoring"


@app.command(
    name="define_host_monitoring_status",
    rich_help_panel=HELP_PANEL,
    hidden=True,
    deprecated=True,
)
@app.command(name="monitor_host", rich_help_panel=HELP_PANEL)
def monitor_host(
    hostname: str = typer.Argument(
        help="Name of host",
        show_default=False,
    ),
    new_status: MonitoringStatus = typer.Argument(
        help="Monitoring status",
        case_sensitive=False,
        show_default=False,
    ),
) -> None:
    """Monitor or unmonitor a host."""
    from zabbix_cli.models import Result

    host = app.state.client.get_host(hostname)
    app.state.client.update_host_status(host, new_status)
    render_result(
        Result(
            message=f"Updated host {hostname!r}. New monitoring status: {new_status}"
        )
    )


@app.command(name="show_host_inventory", rich_help_panel=HELP_PANEL)
def show_host_inventory(
    hostname_or_id: str = typer.Argument(
        help="Hostname or ID",
        show_default=False,
    ),
) -> None:
    """Show host inventory details for a specific host."""
    # TODO: support undocumented filter argument from V2
    # TODO: Add mapping of inventory keys to human readable names (Web GUI names)
    host = app.state.client.get_host(hostname_or_id, select_inventory=True)
    render_result(host.inventory)


@app.command(name="update_host_inventory", rich_help_panel=HELP_PANEL)
def update_host_inventory(
    ctx: typer.Context,
    hostname_or_id: str = typer.Argument(
        help="Hostname or ID of host.",
        show_default=False,
    ),
    key: str = typer.Argument(
        help="Inventory key",
        show_default=False,
    ),
    value: str = typer.Argument(
        help="Inventory value",
        show_default=False,
    ),
) -> None:
    """Update a host inventory field.

    Inventory fields in the API do not always match Web GUI field names.
    Use `zabbix-cli -o json show_host_inventory <hostname>` to see the available fields.
    """
    from zabbix_cli.models import Result

    host = app.state.client.get_host(hostname_or_id)
    to_update = {key: value}
    app.state.client.update_host_inventory(host, to_update)
    render_result(Result(message=f"Updated inventory field {key!r} for host {host}."))



"""Utilities for parsing and validating command-line arguments."""

# NOTE: Should be moved to `commands.common` and merged with the existing `commands.common.args` module.

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Optional

import typer

from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import print_help

if TYPE_CHECKING:
    from zabbix_cli.app import StatefulApp
    from zabbix_cli.pyzabbix.types import Host
    from zabbix_cli.pyzabbix.types import HostGroup
    from zabbix_cli.pyzabbix.types import Template
    from zabbix_cli.pyzabbix.types import TemplateGroup


def is_set(ctx: typer.Context, option: str) -> bool:
    """Check if option is set in context."""
    from click.core import ParameterSource

    src = ctx.get_parameter_source(option)
    if not src:
        logging.warning("Parameter %s not found in context.", option)
        return False

    # HACK: A typer callback that sets an empty list as a default value
    # for a field with a None default and `Optiona[List[str]]` type
    # will cause the parameter to be set to a non-default value,
    # and thus be considered "set" by Click. That is wrong.
    # This is only relevant because of `zabbix_cli._v2_compat.ARGS_POSITIONAL`.
    # It might be better to check for that specific case instead of this
    # general check, which might catch other cases that are not bugs (?)
    if ctx.params.get(option) == [] and option == "args":
        return False

    return src != ParameterSource.DEFAULT


def parse_int_arg(arg: str) -> int:
    """Convert string to int."""
    try:
        return int(arg.strip())
    except ValueError as e:
        raise ZabbixCLIError(f"Invalid integer value: {arg}") from e


def parse_list_arg(arg: Optional[str], *, keep_empty: bool = False) -> list[str]:
    """Convert comma-separated string to list."""
    try:
        args = arg.strip().split(",") if arg else []
        if not keep_empty:
            args = [a for a in args if a]
        return args
    except ValueError as e:
        raise ZabbixCLIError(f"Invalid comma-separated string value: {arg}") from e


def parse_int_list_arg(arg: str) -> list[int]:
    """Convert comma-separated string of ints to list of ints."""
    args = parse_list_arg(
        arg,
        keep_empty=False,  # Important that we never try to parse empty strings as ints
    )
    try:
        return list(map(parse_int_arg, args))
    except ValueError as e:
        raise ZabbixCLIError(f"Invalid comma-separated string value: {arg}") from e


def parse_hostgroups_arg(
    app: StatefulApp,
    hgroup_names_or_ids: Optional[str],
    *,
    strict: bool = False,
    select_hosts: bool = False,
    select_templates: bool = False,
) -> list[HostGroup]:
    """Parse host group names or IDs and return a list of host groups."""
    from zabbix_cli.output.console import exit_err
    from zabbix_cli.output.prompts import str_prompt

    if not hgroup_names_or_ids:
        hgroup_names_or_ids = str_prompt("Host group(s)")

    hg_args = [h.strip() for h in hgroup_names_or_ids.split(",")]
    if not hg_args:
        exit_err("At least one host group name/ID is required.")

    hostgroups = app.state.client.get_hostgroups(
        *hg_args,
        search=True,
        select_hosts=select_hosts,
        select_templates=select_templates,
    )
    if not hostgroups:
        exit_err(f"No host groups found matching {hgroup_names_or_ids}")
    if strict and len(hostgroups) != len(hg_args):
        exit_err(f"Found {len(hostgroups)} host groups, expected {len(hostgroups)}")
    return hostgroups


def parse_hosts_arg(
    app: StatefulApp,
    hostnames_or_ids: Optional[str],
    *,
    strict: bool = False,
) -> list[Host]:
    """Parse host names or IDs and return a list of hosts."""
    from zabbix_cli.output.console import exit_err
    from zabbix_cli.output.prompts import str_prompt

    if not hostnames_or_ids:
        hostnames_or_ids = str_prompt("Host(s)")

    host_args = parse_list_arg(hostnames_or_ids)
    if not host_args:
        exit_err("At least one host name/ID is required.")

    hosts = app.state.client.get_hosts(*host_args, search=True)
    if not hosts:
        exit_err(f"No hosts found matching {hostnames_or_ids}")
    if strict and len(hosts) != len(host_args):
        exit_err(f"Found {len(hosts)} hosts, expected {len(host_args)}")
    return hosts


def parse_templates_arg(
    app: StatefulApp,
    template_names_or_ids: Optional[str],
    *,
    strict: bool = False,
    select_hosts: bool = False,
) -> list[Template]:
    """Parse template names or IDs and return a list of templates."""
    from zabbix_cli.output.console import exit_err

    template_args = parse_list_arg(template_names_or_ids)
    if not template_args:
        exit_err("At least one template name/ID is required.")

    templates = app.state.client.get_templates(
        *template_args, select_hosts=select_hosts
    )
    if not templates:
        exit_err(f"No templates found matching {template_names_or_ids}")
    if strict and len(templates) != len(template_args):
        exit_err(f"Found {len(templates)} templates, expected {len(template_args)}")

    return templates


def parse_templategroups_arg(
    app: StatefulApp,
    tgroup_names_or_ids: str,
    *,
    strict: bool = False,
    select_templates: bool = False,
) -> list[TemplateGroup]:
    """Parse template group names or IDs and return a list of template groups."""
    tg_args = parse_list_arg(tgroup_names_or_ids)
    if not tg_args:
        exit_err("At least one template group name/ID is required.")

    templategroups = app.state.client.get_templategroups(
        *tg_args, search=True, select_templates=select_templates
    )
    if not templategroups:
        exit_err(f"No template groups found matching {tgroup_names_or_ids}")
    if strict and len(templategroups) != len(tg_args):
        exit_err(
            f"Found {len(templategroups)} template groups, expected {len(templategroups)}"
        )
    return templategroups


TRUE_CHOICES = ["true", "y", "yes", "1", "on"]
FALSE_CHOICES = ["false", "n", "no", "0", "off"]


def parse_bool_arg(arg: str) -> bool:
    """Convert string to bool."""
    arg = arg.strip().lower()
    if arg in TRUE_CHOICES:
        return True
    elif arg in FALSE_CHOICES:
        return False
    else:
        raise ZabbixCLIError(f"Invalid boolean value: {arg}")


def parse_path_arg(arg: str, *, must_exist: bool = False) -> Path:
    """Convert string to Path."""
    try:
        p = Path(arg)
    except ValueError as e:
        raise ZabbixCLIError(f"Invalid path value: {arg}") from e
    else:
        if must_exist and not p.exists():
            raise ZabbixCLIError(f"Path does not exist: {arg}")
        return p


### API / arg utilities ###


def get_hostgroup_hosts(
    app: StatefulApp, hostgroups: list[HostGroup] | str
) -> list[Host]:
    """Get all hosts from a list of host groups.

    Args:
        app: The application instance.
        hostgroups: List of host groups or a comma-separated string of host group names.
    """
    if isinstance(hostgroups, str):
        hostgroup_names = parse_list_arg(hostgroups)
        hostgroups = app.state.client.get_hostgroups(
            *hostgroup_names, select_hosts=True, search=True
        )
    # Get all hosts from all host groups
    # Some hosts can be in multiple host groups - ensure no dupes
    hosts: list[Host] = []
    seen: set[str] = set()
    for hg in hostgroups:
        for host in hg.hosts:
            if host.host not in seen:
                hosts.append(host)
                seen.add(host.host)
    return hosts


def check_at_least_one_option_set(ctx: typer.Context) -> None:
    """Check that at least one option is set in the context.

    Useful for commands used to update resources, where all options
    are optional, but at least one is required to make a change.
    """
    optional_params: set[str] = set()
    for param in ctx.command.params:
        if param.required:
            continue
        if not param.name:
            logging.warning("Unnamed parameter in command %s", ctx.command.name)
            continue
        optional_params.add(param.name)
    if not any(is_set(ctx, param) for param in optional_params):
        print_help(ctx)
        exit_err("At least one option is required.")



"""Configuration classes for Zabbix CLI commands."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

from pydantic import AliasChoices
from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from zabbix_cli.config.base import BaseModel
from zabbix_cli.dirs import EXPORT_DIR
from zabbix_cli.pyzabbix.enums import ExportFormat

logger = logging.getLogger(__name__)


class CreateHost(BaseModel):
    """Configuration for the `create_host` command."""

    create_interface: bool = Field(
        default=True,
        description="Create a DNS/IP interface for the host automatically.",
    )
    hostgroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "hostgroups",  # v3.6
            "create_host_hostgroups",  # v3.0
            "create_host_hostgroup",  # v2
        ),
        description="Default host group to add hosts to.",
    )


class CreateNotificationUser(BaseModel):
    """Configuration for the `create_notification_user` command."""

    usergroups: list[str] = Field(
        default=[],
        # Changed in V3: default_notification_users_usergroup -> default_notification_users_usergroups
        validation_alias=AliasChoices(
            "usergroups",  # v3.6
            "default_notification_users_usergroups",  # v3.0
            "default_notification_users_usergroup",  # v2
        ),
        description=(
            "Default user groups to add notification users to when `--usergroups` is not provided."
        ),
    )


class CreateUser(BaseModel):
    """Configuration for the `create_user` command."""

    usergroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "usergroups",  # v3.6
            "default_create_user_usergroups",  # v3.0
            "default_create_user_usergroup",  # v2
        ),
        description=(
            "Default user groups to add users to when `--usergroups` is not provided."
        ),
    )


class _CreateGroupBase(BaseModel):
    ro_groups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "ro_groups",  # v3.6
            "default_create_user_usergroups",  # v3.0
            "default_create_user_usergroup",  # v2
        ),
        description=(
            "Default user groups to give read-only permissions to groups "
            "when `--ro-groups` option is not provided."
        ),
    )
    rw_groups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "rw_groups",  # v3.6
            "default_admin_usergroups",  # v3.0
            "default_admin_usergroup",  # v2
        ),
        description=(
            "Default user groups to give read/write permissions to groups "
            "when `--rw-groups` option is not provided."
        ),
    )

    def __bool__(self) -> bool:
        return bool(self.ro_groups or self.rw_groups)


class CreateHostGroup(_CreateGroupBase):
    """Configuration for the `create_hostgroup` command."""


class CreateTemplateGroup(_CreateGroupBase):
    """Configuration for the `create_templategroup` command."""


class CreateHostOrTemplateGroup(_CreateGroupBase):
    """Shared config for `create_hostgroup` and `create_templategroup` commands.

    Can be used to configure both commands at once. Has no effect if
    `create_hostgroup` or `create_templategroup` is set.
    """


class ExportImport(BaseModel):
    """Shared configuration for `export_configuration` and `import_configuration` commands."""

    directory: Path = Field(
        default=EXPORT_DIR,
        validation_alias=AliasChoices(
            "directory",  # v3.6
            "export_directory",  # v3.0
            "default_directory_exports",  # v2
        ),
        description="Default directory to export files to.",
    )
    format: ExportFormat = Field(
        default=ExportFormat.JSON,
        validation_alias=AliasChoices(
            "format",  # v3.6
            "export_format",  # v3.0
            "default_export_format",  # v2
        ),
        description="Default export format.",
    )
    timestamps: bool = Field(
        default=False,
        # Changed in V3: include_timestamp_export_filename -> export_timestamps
        validation_alias=AliasChoices(
            "timestamps",  # v3.6
            "export_timestamps",  # v3.0
            "include_timestamp_export_filename",  # v2
        ),
        description="Include timestamps in exported filenames.",
    )


# Top-level class composed of individual classes for each command
class CommandConfig(BaseModel):
    """Configuration of commands."""

    # Hosts
    create_host: CreateHost = Field(default_factory=CreateHost)

    # Groups
    create_group: CreateHostOrTemplateGroup = Field(
        default_factory=CreateHostOrTemplateGroup,
    )
    create_hostgroup: CreateHostGroup = Field(default_factory=CreateHostGroup)
    create_templategroup: CreateTemplateGroup = Field(
        default_factory=CreateTemplateGroup,
    )

    # Users
    create_user: CreateUser = Field(default_factory=CreateUser)
    create_notification_user: CreateNotificationUser = Field(
        default_factory=CreateNotificationUser
    )

    # Export
    export: ExportImport = Field(
        default_factory=ExportImport,
        validation_alias=AliasChoices(
            # short-hand + command + combined names
            "export",
            "import",
            "export_configuration",
            "import_configuration",
            "export_import",
        ),
    )

    @model_validator(mode="before")
    @classmethod
    def check_export_import(cls, data: Any) -> Any:
        """Emit warning if multiple export/import aliases are used."""
        if not isinstance(data, dict):
            return data

        if field := cls.model_fields.get("export"):
            if isinstance(field.validation_alias, AliasChoices):
                aliases = field.validation_alias.choices
                # NOTE: we assume alias order is identical to definition order
                # Coerce to str, even though we shouldn't have any AliasPath here
                found = [str(alias) for alias in aliases if alias in data]
                if len(found) > 1:
                    logger.warning(
                        "Multiple export/import configuration sections found (%s). "
                        "Using section: [app.commands.%s]",
                        ", ".join(f"[app.commands.{f}]" for f in found),
                        found[0],
                    )
        return data  # pyright: ignore[reportUnknownVariableType]

    @model_validator(mode="after")
    def check_create_group(self) -> Self:
        """Fallback to `create_group` if `create_{host,template}group` are not set."""
        if (
            # one or both of create_{host,template}group is not set/empty
            any(not f for f in [self.create_hostgroup, self.create_templategroup])
            # Shared config is defined
            and "create_group" in self.model_fields_set
            # and has at least one type of group defined
            and (self.create_group.ro_groups or self.create_group.rw_groups)
        ):
            # Only override if empty/not set
            if not self.create_hostgroup:
                self.create_hostgroup = self.create_hostgroup.model_validate(
                    self.create_group, from_attributes=True
                )
            if not self.create_templategroup:
                self.create_templategroup = self.create_templategroup.model_validate(
                    self.create_group, from_attributes=True
                )
        return self



from __future__ import annotations

import sys
from pathlib import Path

import pytest

sys.path.append(str((Path(__file__) / "../../../docs").resolve()))


# Skip entire test if not found
markup = pytest.importorskip("docs.scripts.utils.markup")
MarkdownSymbol = markup.MarkdownSymbol  # type: ignore


@pytest.mark.parametrize(
    "inp, expect",
    [
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=False,
                code=False,
                codeblock=False,
                end=False,
            ),
            "*",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=True,
                code=False,
                codeblock=False,
                end=False,
            ),
            "**",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=False,
                code=True,
                codeblock=False,
                end=False,
            ),
            "`",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=False,
                code=False,
                codeblock=True,
                end=False,
            ),
            "```\n",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=True,
                code=False,
                codeblock=False,
                end=False,
            ),
            "***",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=False,
                code=True,
                codeblock=False,
                end=False,
            ),
            "*`",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=False,
                code=False,
                codeblock=True,
                end=False,
            ),
            "```\n",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=True,
                code=True,
                codeblock=False,
                end=False,
            ),
            "**`",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=True,
                code=False,
                codeblock=True,
                end=False,
            ),
            "```\n",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=False,
                bold=False,
                code=True,
                codeblock=True,
                end=False,
            ),
            "```\n",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=True,
                code=True,
                codeblock=False,
                end=False,
            ),
            "***`",
        ),
        (
            MarkdownSymbol(
                position=0,
                italic=True,
                bold=True,
                code=False,
                codeblock=True,
                end=False,
            ),
            "```\n",
        ),
        (
            MarkdownSymbol(
                position=0, italic=True, bold=True, code=True, codeblock=True, end=False
            ),
            "```\n",
        ),
    ],
)
def test_markdownsymbol_symbol(
    inp: MarkdownSymbol,  # type: ignore
    expect: str,
) -> None:
    assert inp.symbol == expect  # type: ignore


@pytest.mark.parametrize(
    "inp,expect",
    [
        pytest.param(
            "[example]zabbix-cli --version[/]",
            "```\nzabbix-cli --version\n```",
            id="Example (short end)",
        ),
        pytest.param(
            "[example]zabbix-cli --version[/example]",
            "```\nzabbix-cli --version\n```",  # Same result as above
            id="Example (explicit end style)",
        ),
        pytest.param(
            "[bold example]zabbix-cli --version[/bold example]",
            "```\nzabbix-cli --version\n```",  # Ignoring bold
            id="Example + bold",
        ),
        pytest.param(
            "[example python]zabbix-cli --version[/example python]",
            "```py\nzabbix-cli --version\n```",  # Adding language
            id="Example with language",
        ),
        pytest.param(
            "     [example]zabbix-cli --version[/example]",
            "```\nzabbix-cli --version\n```",  # Ignoring leading spaces
            id="Leading spaces outside style",
        ),
        pytest.param(
            "[example]      zabbix-cli --version[/example]",
            "```\nzabbix-cli --version\n```",  # Ignoring leading spaces inside
            id="Leading spaces inside style",
        ),
    ],
)
def test_markup_to_markdown(inp: str, expect: str) -> None:
    assert markup.markup_to_markdown(inp) == expect  # type: ignore


@pytest.mark.parametrize(
    "style",
    markup.CODE_STYLES,
)
def test_markup_to_markdown_code_styles(style: str) -> None:
    expect = "`--version`"
    assert markup.markup_to_markdown(f"[{style}]--version[/]") == expect  # type: ignore

    # Adding bold or italic doesnt change anything
    # code takes precedence
    assert markup.markup_to_markdown(f"[bold {style}]--version[/]") == expect
    assert markup.markup_to_markdown(f"[italic {style}]--version[/]") == expect
    assert markup.markup_to_markdown(f"[bold italic {style}]--version[/]") == expect


@pytest.mark.parametrize(
    "style",
    markup.CODEBLOCK_STYLES,
)
def test_markup_to_markdown_codeblock_styles(style: str) -> None:
    text = "zabbix-cli remove_hostgroup foo"
    expect = f"```\n{text}\n```"
    assert markup.markup_to_markdown(f"[{style}]{text}[/]") == expect  # type: ignore

    # Adding bold or italic doesnt change anything
    # code takes precedence
    assert markup.markup_to_markdown(f"[italic {style}]{text}[/]") == expect
    assert markup.markup_to_markdown(f"[bold {style}]{text}[/]") == expect
    assert markup.markup_to_markdown(f"[bold {style} italic]{text}[/]") == expect


# TODO: test the example spans of all commands and ensure they render as codeblocks



"""Completer functionality for Click commands in the REPL."""

from __future__ import annotations

import os
import shlex
from collections.abc import Generator
from glob import iglob
from typing import Any
from typing import Optional

import click
from prompt_toolkit.completion import CompleteEvent
from prompt_toolkit.completion import Completer
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document

from zabbix_cli.commands.common.args import CommandParam

__all__ = ["ClickCompleter"]

IS_WINDOWS = os.name == "nt"


AUTO_COMPLETION_PARAM = "shell_complete"


def split_arg_string(string: str, *, posix: bool = True) -> list[str]:
    r"""Split an argument in a shlex-like way, but don't fail if the string is incomplete.

    Args:
        string (str): String to split
        posix (bool, optional): Posix-like parsing. Defaults to True.

    Returns:
        list[str]: String split into tokens.

    Examples:
        >>> split_arg_string("example 'my file")
        ["example", "my file"]
        >>> split_arg_string("example my\\")
        ["example", "my"]
    """
    lex = shlex.shlex(string, posix=posix)
    lex.whitespace_split = True
    lex.commenters = ""
    out: list[str] = []

    try:
        for token in lex:
            out.append(token)
    except ValueError:
        # Raised when end-of-string is reached in an invalid state. Use
        # the partial token as-is. The quote or escape character is in
        # lex.state, not lex.token.
        out.append(lex.token)

    return out


def _resolve_context(args: list[Any], ctx: click.Context) -> click.Context:
    """Resolve context and return the last context in the chain.

    Produce the context hierarchy starting with the command and
    traversing the complete arguments. This only follows the commands,
    it doesn't trigger input prompts or callbacks.


    Args:
        args (list[Any]): List of complete args before the incomplete value.
        ctx (click.Context): `click.Context` object of the CLI group

    Returns:
        click.Context: Resolved context.
    """
    while args:
        command = ctx.command

        if isinstance(command, click.MultiCommand):
            if not command.chain:
                name, cmd, args = command.resolve_command(ctx, args)

                if cmd is None:
                    return ctx

                ctx = cmd.make_context(name, args, parent=ctx, resilient_parsing=True)
                args = ctx.protected_args + ctx.args
            else:
                sub_ctx = ctx
                while args:
                    name, cmd, args = command.resolve_command(ctx, args)

                    if cmd is None:
                        return ctx

                    sub_ctx = cmd.make_context(
                        name,
                        args,
                        parent=ctx,
                        allow_extra_args=True,
                        allow_interspersed_args=False,
                        resilient_parsing=True,
                    )
                    args = sub_ctx.args
                ctx = sub_ctx
                args = [*sub_ctx.protected_args, *sub_ctx.args]
        else:
            break

    return ctx


class ClickCompleter(Completer):
    """Completer for Click commands."""

    __slots__ = ("cli", "ctx", "parsed_args", "parsed_ctx", "ctx_command")

    def __init__(
        self,
        cli: click.Group,
        ctx: click.Context,
        *,
        show_only_unused: bool = False,
        shortest_only: bool = False,
    ) -> None:
        self.cli = cli
        self.ctx = ctx
        self.parsed_args = []
        self.parsed_ctx = ctx
        self.ctx_command = ctx.command
        self.show_only_unused = show_only_unused
        self.shortest_only = shortest_only

    def _get_completion_from_autocompletion_functions(
        self,
        param: click.Parameter,
        autocomplete_ctx: click.Context,
        args: list[str],
        incomplete: str,
    ):
        param_choices: list[Completion] = []
        autocompletions = param.shell_complete(autocomplete_ctx, incomplete)
        for autocomplete in autocompletions:
            param_choices.append(Completion(str(autocomplete.value), -len(incomplete)))
        return param_choices

    def _get_completion_for_Path_types(
        self, param: click.Parameter, args: list[str], incomplete: str
    ) -> list[Completion]:
        if "*" in incomplete:
            return []

        choices: list[Completion] = []
        _incomplete = os.path.expandvars(incomplete)
        search_pattern = _incomplete.strip("'\"\t\n\r\v ").replace("\\\\", "\\") + "*"
        quote = ""

        if " " in _incomplete:
            for i in incomplete:
                if i in ("'", '"'):
                    quote = i
                    break

        for path in iglob(search_pattern):
            if " " in path:
                if quote:
                    path = quote + path
                else:
                    if IS_WINDOWS:
                        path = repr(path).replace("\\\\", "\\")
            else:
                if IS_WINDOWS:
                    path = path.replace("\\", "\\\\")

            choices.append(
                Completion(
                    str(path),
                    -len(incomplete),
                    display=str(os.path.basename(path.strip("'\""))),
                )
            )

        return choices

    def _get_completion_for_Boolean_type(self, param: click.Parameter, incomplete: str):
        return [
            Completion(str(k), -len(incomplete), display_meta=str("/".join(v)))
            for k, v in {
                "true": ("1", "true", "t", "yes", "y", "on"),
                "false": ("0", "false", "f", "no", "n", "off"),
            }.items()
            if any(i.startswith(incomplete) for i in v)
        ]

    def _get_completion_from_command_param(
        self, param: click.Parameter, incomplete: str
    ) -> list[Completion]:
        return [
            Completion(command, -len(incomplete))
            for command in self.cli.list_commands(self.parsed_ctx)
            if command.startswith(incomplete)
        ]

    def _get_completion_from_params(
        self,
        autocomplete_ctx: click.Context,
        args: list[str],
        param: click.Parameter,
        incomplete: str,
    ) -> list[Completion]:
        choices: list[Completion] = []
        if isinstance(param.type, click.types.BoolParamType):
            # Only suggest completion if parameter is not a flag
            if isinstance(param, click.Option) and not param.is_flag:
                choices.extend(self._get_completion_for_Boolean_type(param, incomplete))
        elif isinstance(param.type, (click.Path, click.File)):
            choices.extend(self._get_completion_for_Path_types(param, args, incomplete))
        elif isinstance(param.type, CommandParam):
            choices.extend(self._get_completion_from_command_param(param, incomplete))
        elif getattr(param, AUTO_COMPLETION_PARAM, None) is not None:
            choices.extend(
                self._get_completion_from_autocompletion_functions(
                    param,
                    autocomplete_ctx,
                    args,
                    incomplete,
                )
            )

        return choices

    def _get_completion_for_cmd_args(
        self,
        ctx_command: click.Command,
        incomplete: str,
        autocomplete_ctx: click.Context,
        args: list[str],
    ) -> list[Completion]:
        choices: list[Completion] = []
        param_called = False

        for param in ctx_command.params:
            if isinstance(param.type, click.types.UnprocessedParamType):
                return []

            elif getattr(param, "hidden", False):
                continue

            elif isinstance(param, click.Option):
                opts = param.opts + param.secondary_opts
                previous_args = args[: param.nargs * -1]
                current_args = args[param.nargs * -1 :]

                # Show only unused opts
                already_present = any(opt in previous_args for opt in opts)
                hide = self.show_only_unused and already_present and not param.multiple

                # Show only shortest opt
                if (
                    self.shortest_only
                    and not incomplete  # just typed a space
                    # not selecting a value for a longer version of this option
                    and args[-1] not in opts
                ):
                    opts = [min(opts, key=len)]

                for option in opts:
                    # We want to make sure if this parameter was called
                    # If we are inside a parameter that was called, we want to show only
                    # relevant choices
                    if option in current_args:  # noqa: E203
                        param_called = True
                        break

                    elif option.startswith(incomplete) and not hide:
                        choices.append(
                            Completion(
                                str(option),
                                -len(incomplete),
                                display_meta=str(param.help or ""),
                            )
                        )

                if param_called:
                    choices = self._get_completion_from_params(
                        autocomplete_ctx, args, param, incomplete
                    )
                    break

            elif isinstance(param, click.Argument):
                choices.extend(
                    self._get_completion_from_params(
                        autocomplete_ctx, args, param, incomplete
                    )
                )

        return choices

    def get_completions(
        self, document: Document, complete_event: Optional[CompleteEvent] = None
    ) -> Generator[Completion, Any, None]:
        """Generator of completions for the current input.

        Args:
            document (Document): Current prompt_toolkit document.
            complete_event (Optional[CompleteEvent], optional): Event that triggered the completion. Defaults to None.

        Yields:
            Generator[Completion, Any, None]: Completions for the current input.
        """
        # Code analogous to click._bashcomplete.do_complete

        args = split_arg_string(document.text_before_cursor, posix=False)

        choices: list[Completion] = []
        cursor_within_command = (
            document.text_before_cursor.rstrip() == document.text_before_cursor
        )

        if document.text_before_cursor.startswith(("!", ":")):
            return

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ""

        if self.parsed_args != args:
            self.parsed_args = args
            try:
                self.parsed_ctx = _resolve_context(args, self.ctx)
            except Exception:
                return  # autocompletion for nonexistent cmd can throw here
            self.ctx_command = self.parsed_ctx.command

        if getattr(self.ctx_command, "hidden", False):
            return

        try:
            choices.extend(
                self._get_completion_for_cmd_args(
                    self.ctx_command, incomplete, self.parsed_ctx, args
                )
            )

            if isinstance(self.ctx_command, click.MultiCommand):
                incomplete_lower = incomplete.lower()

                for name in self.ctx_command.list_commands(self.parsed_ctx):
                    command = self.ctx_command.get_command(self.parsed_ctx, name)
                    if getattr(command, "hidden", False):
                        continue

                    elif name.lower().startswith(incomplete_lower):
                        choices.append(
                            Completion(
                                str(name),
                                -len(incomplete),
                                display_meta=getattr(command, "short_help", ""),
                            )
                        )

        except Exception as e:
            click.echo(f"{type(e).__name__}: {str(e)}")

        # If we are inside a parameter that was called, we want to show only
        # relevant choices
        # if param_called:
        #     choices = param_choices

        yield from choices



from __future__ import annotations

from typing import Optional

import typer

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import OPTION_LIMIT
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.render import render_result
from zabbix_cli.utils.args import parse_list_arg


@app.command(
    name="show_last_values",
    rich_help_panel="Host Monitoring",  # Moved to host monitoring for now
    examples=[
        Example(
            "Get items starting with 'MongoDB'",
            "show_last_values 'MongoDB*'",
        ),
        Example(
            "Get items containing 'memory'",
            "show_last_values '*memory*'",
        ),
        Example(
            "Get all items (WARNING: slow!)",
            "show_last_values '*'",
        ),
    ],
)
def show_last_values(
    ctx: typer.Context,
    item: str = typer.Argument(
        help="Item names or IDs. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    group: bool = typer.Option(
        False, "--group", help="Group items with the same value."
    ),
    limit: Optional[int] = OPTION_LIMIT,
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Show the last values of given items of monitored hosts."""
    from zabbix_cli.commands.results.item import ItemResult
    from zabbix_cli.commands.results.item import group_items
    from zabbix_cli.models import AggregateResult

    if args:
        if not len(args) == 1:
            exit_err("Invalid number of positional arguments. Use options instead.")
        group = args[0] == "1"
        # No format arg in V2...

    names_or_ids = parse_list_arg(item)
    with app.status("Fetching items..."):
        items = app.state.client.get_items(
            *names_or_ids, select_hosts=True, monitored=True, limit=limit
        )

    # HACK: not super elegant, but this allows us to match V2 output while
    # with and without the --group flag, as well as ALSO rendering the entire
    # Item object instead of just a subset of fields.
    # Ideally, it would be nice to not have to re-validate when not grouping
    # but I'm not sure how to do that in Pydantic V2?
    if group:
        res = group_items(items)
        render_result(AggregateResult(result=res))
    else:
        res = [ItemResult.from_item(item) for item in items]
    render_result(AggregateResult(result=res))



from __future__ import annotations

import json
from contextlib import nullcontext
from typing import TYPE_CHECKING
from typing import Any

import typer

from zabbix_cli.output.console import console
from zabbix_cli.output.console import error
from zabbix_cli.output.console import success
from zabbix_cli.state import get_state

if TYPE_CHECKING:
    from pydantic import BaseModel

    from zabbix_cli.models import BaseResult
    from zabbix_cli.models import TableRenderable


def wrap_result(result: BaseModel) -> BaseResult:
    """Wraps a BaseModel instance in a Result object so that it receives
    `return_code`, `errors`, and `message` fields, with the original object
    is available as `result`.

    Does nothing if the function argument is already a BaseResult instance.
    """
    from zabbix_cli.models import BaseResult
    from zabbix_cli.models import Result

    if isinstance(result, BaseResult):
        return result
    # TODO: handle AggregateResult? (8 months later: What did I mean by this?)
    return Result(result=result)


def render_result(
    result: TableRenderable,
    ctx: typer.Context | None = None,
    **kwargs: Any,
) -> None:
    """Render the result of a command stdout or file.

    Parameters
    ----------
    result: TableRenderable,
        The result of a command. All commands produce a TableRenderable (BaseModel).
    ctx : typer.Context, optional
        The typer context from the command invocation, by default None
    **kwargs
        Additional keyword arguments to pass to the render function.
    """
    from zabbix_cli.config.constants import OutputFormat

    # Short form aliases
    state = get_state()
    fmt = state.config.app.output.format
    # paging = state.config.output.paging
    paging = False  # TODO: implement

    ctx_manager = console.pager() if paging else nullcontext()
    with ctx_manager:
        if fmt == OutputFormat.JSON:
            if state.config.app.legacy_json_format:
                render_json_legacy(result, ctx, **kwargs)
            else:
                render_json(result, ctx, **kwargs)
        elif fmt == OutputFormat.TABLE:
            render_table(result, ctx, **kwargs)
        # TODO: implement CSV
        else:
            raise ValueError(f"Unknown output format {fmt!r}.")


def render_table(
    result: TableRenderable, ctx: typer.Context | None = None, **kwargs: Any
) -> None:
    """Render the result of a command as a table if possible.
    If result contains a message, print success message instead.
    """
    # TODO: be able to print message _AND_ table
    # The Result/TableRenderable dichotomy is a bit of a mess
    from zabbix_cli.models import Result
    from zabbix_cli.models import ReturnCode

    if isinstance(result, Result) and result.message:
        if result.return_code == ReturnCode.ERROR:
            error(result.message)
        else:
            success(result.message)
    else:
        tbl = result.as_table()
        if not tbl.rows:
            if not result.empty_ok:
                console.print("No results found.")
        else:
            console.print(tbl)


def render_json(
    result: TableRenderable,
    ctx: typer.Context | None = None,
    **kwargs: Any,
) -> None:
    """Render the result of a command as JSON."""
    from zabbix_cli.models import ReturnCode

    result = wrap_result(result)
    o_json = result.model_dump_json(indent=2, by_alias=True)
    console.print_json(o_json, indent=2, sort_keys=False)
    if result.message:
        if result.return_code == ReturnCode.ERROR:
            error(result.message)
        else:
            success(result.message)


def render_json_legacy(
    result: TableRenderable,
    ctx: typer.Context | None = None,
    **kwargs: Any,
) -> None:
    """Render the result of a command as JSON (legacy V2 format).

    Result is always a dict with numeric string keys.

    Note:
    ----
    This function is very hacky, and will inevitably contain a number of band-aid
    fixes to enable 1:1 compatibility with the legacy V2 JSON format.
    We should try to move away from this format ASAP, so we can remove
    this function and all its hacks.
    """
    from zabbix_cli.models import Result

    # If we have a message, it should not be indexed
    # NOTE: do we have a more accurate heuristic for this?
    if isinstance(result, Result) and result.message:
        j = result.model_dump_json(indent=2)
    else:
        from zabbix_cli.models import AggregateResult

        jdict: dict[str, Any] = {}  # always a dict in legacy mode
        res = result.model_dump(mode="json", by_alias=True)
        if isinstance(result, AggregateResult):
            py_result = res.get("result", [])
        else:
            py_result = [res]

        for idx, item in enumerate(py_result):
            jdict[str(idx)] = item
        j = json.dumps(jdict, indent=2)
    console.print_json(j, indent=2, sort_keys=False)



# Rich markup styles for the CLI
from __future__ import annotations

from typing import Any

from rich.theme import Theme
from strenum import StrEnum
from typer.rich_utils import STYLE_OPTION

# NOTE: we define these enums to allow us to parse the markup text and
#       correctly convert it to markdown in the docs. Without this, we would
#       have to hard-code each style to correspond to a specific markdown formatting
#       in the docs generator, which would be error-prone and difficult to maintain.
#       E.g. [command]zabbix-cli hostgroup_remove foo[/] becomes `zabbix-cli hostgroup_remove foo`
#       while [example]zabbix-cli --version[/] becomes ```\nzabbix-cli --version\n``` (code block)


class CodeBlockStyle(StrEnum):
    """Names of styles for text representing code blocks.

    Displayed as a code block in markdown.
    """

    EXAMPLE = "example"
    """An example command."""

    # TODO: add language style here or as separate enum? if so, how to parse in docs?

    # NOTE: add "code" style here or in CodeStyle?


class CodeStyle(StrEnum):
    """Names of styles for text representing code, configuration or commands.

    Displayed as inline code-formatted text in markdown.
    """

    CONFIG_OPTION = "configopt"
    """Configuration file option/key/entry."""

    CLI_OPTION = "option"
    """CLI option, e.g. --verbose."""

    CLI_VALUE = "value"
    """CLI value, arg or metavar e.g. 'FILE'."""

    CLI_COMMAND = "command"
    """CLI command e.g. 'hostgroup_remove'."""

    CODE = "code"


class TextStyle(StrEnum):
    """Names of styles for non-code text"""

    WARNING = "warning"
    ERROR = "error"
    INFO = "info"
    SUCCESS = "success"


class TableStyle(StrEnum):
    """Names of styles for table headers, rows, etc."""

    HEADER = "table_header"


####################
# Colors
####################
# Colors should be used to colorize output and help define styles,
# but they should not contain any formatting (e.g. bold, italic, `x` on `y`, etc.)
####################


# TODO: refactor and define info, success, warning, error as STYLES, not COLORS.
#       Having multiple members with the same value is bad
class Color(StrEnum):
    INFO = "default"
    SUCCESS = "green"
    WARNING = "yellow"
    ERROR = "red"
    YELLOW = "yellow"
    GREEN = "green"
    RED = "red"
    MAGENTA = "magenta"
    CYAN = "cyan"
    BLUE = "blue"

    def __call__(self, message: str) -> str:
        return f"[{self.value}]{message}[/]"


# TODO: define primary, secondary, tertiary colors, then replace
#       green with primary, magenta with secondary, yellow with tertiary, etc.
RICH_THEME = Theme(
    {
        CodeBlockStyle.EXAMPLE.value: "bold green",
        CodeStyle.CLI_COMMAND.value: "bold green",
        CodeStyle.CLI_OPTION.value: STYLE_OPTION,
        CodeStyle.CLI_VALUE.value: "bold magenta",
        CodeStyle.CONFIG_OPTION.value: "italic yellow",
        CodeStyle.CODE.value: "bold green",
        TextStyle.SUCCESS.value: Color.SUCCESS,
        TextStyle.WARNING.value: f"bold {Color.WARNING}",
        TextStyle.ERROR.value: f"bold {Color.ERROR}",
        TextStyle.INFO.value: Color.SUCCESS,
        TableStyle.HEADER.value: "bold green",
    }
)

# NOTE: this seems a bit TOO decoupled? Would be nice if we could define
# styles as a combination of style name + style value.
# That would require rewriting parts of the docs utils for parsing Rich
# markup to markdown.


def blue(message: str) -> str:
    return f"[blue]{message}[/]"


def cyan(message: str) -> str:
    return f"[cyan]{message}[/]"


def green(message: str) -> str:
    return f"[green]{message}[/]"


def magenta(message: str) -> str:
    return f"[magenta]{message}[/]"


def red(message: str) -> str:
    return f"[red]{message}[/]"


def yellow(message: str) -> str:
    return f"[yellow]{message}[/]"


def bold(message: str) -> str:
    return f"[bold]{message}[/]"


def warning(message: str) -> str:
    return f"[warning]{message}[/]"


def error(message: str) -> str:
    return f"[error]{message}[/]"


def success(message: str) -> str:
    return f"[success]{message}[/]"


def info(message: str) -> str:
    return f"[info]{message}[/]"


####################
# Emojis
####################


EMOJI_YES = ":white_check_mark:"
EMOJI_NO = ":cross_mark:"


class Icon(StrEnum):
    DEBUG = "⚙"
    INFO = "!"
    OK = "✓"
    ERROR = "✗"
    PROMPT = "?"
    WARNING = "⚠"


# TODO: replace all use of constants with these enums
class Emoji(StrEnum):
    YES = EMOJI_YES
    NO = EMOJI_NO

    @classmethod
    def fmt_bool(cls, value: bool) -> str:  # noqa: FBT001
        return success(cls.YES) if value else error(cls.NO)


def render_config_option(option: str) -> str:
    """Render a configuration file option/key/entry."""
    return f"[{CodeStyle.CONFIG_OPTION}]{option}[/]"


def render_cli_option(option: str) -> str:
    """Render a CLI option."""
    return f"[{CodeStyle.CLI_OPTION}]{option}[/]"


def render_cli_value(value: Any) -> str:
    """Render a CLI value/argument."""
    return f"[{CodeStyle.CLI_VALUE}]{value!r}[/]"


def render_cli_command(value: str) -> str:
    """Render a CLI command."""
    return f"[{CodeStyle.CLI_COMMAND}]{value}[/]"



from __future__ import annotations

from inline_snapshot import snapshot
from zabbix_cli.repl.repl import _help_internal  # pyright: ignore[reportPrivateUsage]


def test_help_internal() -> None:
    assert _help_internal() == snapshot(
        """\
REPL help:

  External Commands:
    prefix external commands with "!"

  Internal Commands:
    prefix internal commands with ":"
    :exit, :q, :quit  exits the repl
    :?, :h, :help     displays general help information
"""
    )



"""Models for rendering results of commands.

Should not be imported on startup, as we don't want to build Pydantic models
until we actually need them - this has a massive startup time impact.

Each command module should have a corresponding module in this package that
defines the models for its results. i.e. `zabbix_cli.commands.host` should
define its result models in  `zabbix_cli.commands.results.host`.
"""

from __future__ import annotations



# Run the config module with python -m zabbix_cli.config
from __future__ import annotations

if __name__ == "__main__":
    import typer

    from zabbix_cli.config.run import main

    typer.run(main)



"""Compatibility functions to support different Zabbix API versions."""

from __future__ import annotations

from typing import Literal

from packaging.version import Version

# TODO (pederhan): rewrite these functions as some sort of declarative data
# structure that can be used to determine correct parameters based on version
# if we end up with a lot of these functions. For now, this is fine.
# OR we could turn it into a mixin class?

# Compatibility methods for Zabbix API objects properties and method parameters.
# Returns the appropriate property name for the given Zabbix version.
#
# FORMAT: <object>_<property>
# EXAMPLE: user_name() (User object, name property)
#
# NOTE: All functions follow the same pattern:
# Early return if the version is older than the version where the property
# was deprecated, otherwise return the new property name as the default.


def host_proxyid(version: Version) -> Literal["proxy_hostid", "proxyid"]:
    # https://support.zabbix.com/browse/ZBXNEXT-8500
    # https://www.zabbix.com/documentation/7.0/en/manual/api/changes#host
    if version.release < (7, 0, 0):
        return "proxy_hostid"
    return "proxyid"


def host_available(version: Version) -> Literal["available", "active_available"]:
    # TODO: find out why this was changed and what it signifies
    # NO URL YET
    if version.release < (6, 4, 0):
        return "available"  # unsupported in < 6.4.0
    return "active_available"


def login_user_name(version: Version) -> Literal["user", "username"]:
    # https://support.zabbix.com/browse/ZBXNEXT-8085
    # Deprecated in 5.4.0, removed in 6.4.0
    # login uses different parameter names than the User object before 6.4
    # From 6.4 and onwards, login and user.<method> use the same parameter names
    # See: user_name
    if version.release < (5, 4, 0):
        return "user"
    return "username"


def proxy_name(version: Version) -> Literal["host", "name"]:
    # https://support.zabbix.com/browse/ZBXNEXT-8500
    # https://www.zabbix.com/documentation/7.0/en/manual/api/changes#proxy
    if version.release < (7, 0, 0):
        return "host"
    return "name"


def role_id(version: Version) -> Literal["roleid", "type"]:
    # https://support.zabbix.com/browse/ZBXNEXT-6148
    # https://www.zabbix.com/documentation/5.2/en/manual/api/changes_5.0_-_5.2#role
    if version.release < (5, 2, 0):
        return "type"
    return "roleid"


def user_name(version: Version) -> Literal["alias", "username"]:
    # https://support.zabbix.com/browse/ZBXNEXT-8085
    # Deprecated in 5.4, removed in 6.4
    # However, historically we have used "alias" as the parameter name
    # pre-6.0.0, so we maintain that behavior here
    if version.release < (6, 0, 0):
        return "alias"
    return "username"


def user_medias(version: Version) -> Literal["user_medias", "medias"]:
    # https://support.zabbix.com/browse/ZBX-17955
    # Deprecated in 5.2, removed in 6.4
    if version.release < (5, 2, 0):
        return "user_medias"
    return "medias"


def usergroup_hostgroup_rights(
    version: Version,
) -> Literal["rights", "hostgroup_rights"]:
    # https://support.zabbix.com/browse/ZBXNEXT-2592
    # https://www.zabbix.com/documentation/6.2/en/manual/api/changes_6.0_-_6.2
    # Deprecated in 6.2
    if version.release < (6, 2, 0):
        return "rights"
    return "hostgroup_rights"


# NOTE: can we remove this function? Or are we planning on using it to
# assign rights for templates?
def usergroup_templategroup_rights(
    version: Version,
) -> Literal["rights", "templategroup_rights"]:
    # https://support.zabbix.com/browse/ZBXNEXT-2592
    # https://www.zabbix.com/documentation/6.2/en/manual/api/changes_6.0_-_6.2
    # Deprecated in 6.2
    if version.release < (6, 2, 0):
        return "rights"
    return "templategroup_rights"


### API params
# API parameter functions are in the following format:
# param_<object>_<method>_<param>
# So to get the "groups" parameter for the "host.get" method, you would call:
# param_host_get_groups()


def param_host_get_groups(
    version: Version,
) -> Literal["selectHostGroups", "selectGroups"]:
    # https://support.zabbix.com/browse/ZBXNEXT-2592
    # hhttps://www.zabbix.com/documentation/6.2/en/manual/api/changes_6.0_-_6.2#host
    if version.release < (6, 2, 0):
        return "selectGroups"
    return "selectHostGroups"


def param_maintenance_create_groupids(
    version: Version,
) -> Literal["groupids", "groups"]:
    # https://support.zabbix.com/browse/ZBXNEXT-2592
    # https://www.zabbix.com/documentation/6.2/en/manual/api/changes_6.0_-_6.2#host
    if version.release < (6, 2, 0):
        return "groups"
    return "groupids"



"""Commands that interact with the application itself."""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Optional

import typer
from click import Command

from zabbix_cli.app import app
from zabbix_cli.commands.common.args import OPTION_LIMIT
from zabbix_cli.config.constants import SecretMode
from zabbix_cli.dirs import CONFIG_DIR
from zabbix_cli.dirs import DATA_DIR
from zabbix_cli.dirs import EXPORT_DIR
from zabbix_cli.dirs import LOGS_DIR
from zabbix_cli.dirs import SITE_CONFIG_DIR
from zabbix_cli.exceptions import ConfigExistsError
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import print_path
from zabbix_cli.output.console import print_toml
from zabbix_cli.output.console import success
from zabbix_cli.output.render import render_result
from zabbix_cli.utils.fs import open_directory

if TYPE_CHECKING:
    from zabbix_cli.config.model import Config


HELP_PANEL = "CLI"


class DirectoryType(Enum):
    """Directory types."""

    CONFIG = "config"
    DATA = "data"
    LOGS = "logs"
    SITE_CONFIG = "siteconfig"
    EXPORTS = "exports"

    def as_path(self) -> Path:
        DIR_MAP = {
            DirectoryType.CONFIG: CONFIG_DIR,
            DirectoryType.DATA: DATA_DIR,
            DirectoryType.LOGS: LOGS_DIR,
            DirectoryType.SITE_CONFIG: SITE_CONFIG_DIR,
            DirectoryType.EXPORTS: EXPORT_DIR,
        }
        d = DIR_MAP.get(self)
        if d is None:
            raise ZabbixCLIError(f"No default path available for {self!r}.")
        return d


def get_directory(directory_type: DirectoryType, config: Optional[Config]) -> Path:
    if config:
        if directory_type == DirectoryType.CONFIG and config.config_path:
            return config.config_path.parent
        elif directory_type == DirectoryType.EXPORTS:
            return config.app.export_directory
    return directory_type.as_path()


@app.command("debug", hidden=True, rich_help_panel=HELP_PANEL)
def debug_cmd(
    ctx: typer.Context,
    with_auth: bool = typer.Option(
        False, "--auth", help="Include auth token in the result."
    ),
) -> None:
    """Print debug info."""
    from zabbix_cli.commands.results.cli import DebugInfo

    render_result(DebugInfo.from_debug_data(app.state, with_auth=with_auth))


@app.command("help", rich_help_panel=HELP_PANEL)
def help(
    ctx: typer.Context,
    command: Optional[Command] = typer.Argument(None, help="Command name"),
) -> None:
    """Show help for a commmand"""
    # TODO: patch get_help() to make it return a string instead of magically
    # printing to stdout, which we have no control over.
    if not command:
        ctx.find_root().get_help()
        return

    # HACK: Set the info name to the resolved command name, otherwise
    # when we call get_help, it will use the name of the help command
    # instead of the resolved command name.
    # Maybe we can use make_context for this?
    ctx.info_name = command.name
    command.get_help(ctx)


@app.command("init", rich_help_panel=HELP_PANEL)
def init(
    ctx: typer.Context,
    config_file: Optional[Path] = typer.Option(
        None, "--config-file", "-c", help="Location of the config file."
    ),
    overwrite: bool = typer.Option(
        False, "--overwrite", help="Overwrite existing config"
    ),
    url: Optional[str] = typer.Option(
        None, "--url", "-u", help="Zabbix API URL to use."
    ),
    wizard: bool = typer.Option(
        True,
        "--wizard/--no-wizard",
        help="Run the interactive wizard to set up the configuration file.",
    ),
) -> None:
    """Create and initialize config file."""
    from zabbix_cli.commands.cli_configwizard import run_wizard
    from zabbix_cli.config.constants import DEFAULT_CONFIG_FILE
    from zabbix_cli.config.utils import init_config

    if config_file is None:
        config_file = app.state.config.config_path

    try:
        config = init_config(
            config=app.state.config,
            config_file=config_file,
            overwrite=overwrite,
            url=url,
        )
    except ConfigExistsError as e:
        raise ZabbixCLIError(f"{e}. Use [option]--overwrite[/] to overwrite it") from e

    if not config.config_path:
        # If the config file was not specified, set it to the default
        config.config_path = DEFAULT_CONFIG_FILE

    if wizard:
        run_wizard(config)

    config.dump_to_file(config.config_path)
    info(f"Configuration file created: {config.config_path}")
    app.state.config = config


@app.command(name="login", rich_help_panel=HELP_PANEL)
def login(
    ctx: typer.Context,
    username: Optional[str] = typer.Option(
        None,
        "--username",
        "-u",
        help="Username to log in with.",
        show_default=False,
    ),
    password: Optional[str] = typer.Option(
        None,
        "--password",
        "-p",
        help="Password to log in with.",
        show_default=False,
    ),
    token: Optional[str] = typer.Option(
        None,
        "--token",
        "-t",
        help="API token to log in with.",
        show_default=False,
    ),
) -> None:
    """Trigger a new login prompt in the REPL.

    Ends the current session before triggering a new login prompt.

    Triggers a login prompt if [option]--username[/] & [option]--password[/] or [option]--token[/]
    is not provided.


    Creates a new auth token file if enabled in the config.
    """
    from pydantic import SecretStr

    from zabbix_cli.auth import Authenticator

    if not app.state.repl:
        raise ZabbixCLIError("This command is only available in the REPL.")

    # Copy config to avoid invalid/incomplete state if login fails
    config = app.state.config.model_copy(deep=True)

    # Set auth info in the config to use as defaults in prompts
    config.api.username = username or ""
    config.api.auth_token = SecretStr(token or "")
    config.api.password = SecretStr(password or "")

    # Try to end current session
    # We might not have an active session, which means any attempt to
    # terminate it will result in an error. Catch and log it.
    try:
        app.state.logout()
    except ZabbixCLIError as e:
        app.logger.warning(f"Failed to terminate session in login command: {e}")

    if username and password:
        client, info = Authenticator.login_with_username_password(
            config, username, password
        )
        success(f"Logged in to {config.api.url} as {info.credentials.username}.")
    elif token:
        client, info = Authenticator.login_with_token(config, token=token)
        success(f"Logged in to {config.api.url} with token.")
    else:
        client, info = Authenticator.login_with_prompt(config)
        success(f"Logged in to {config.api.url} as {info.credentials.username}.")

    app.state.client = client
    app.state.config = config


@app.command("migrate_config", rich_help_panel=HELP_PANEL)
def migrate_config(
    ctx: typer.Context,
    source: Optional[Path] = typer.Option(
        None, "--source", "-s", help="Location of the config file to migrate."
    ),
    destination: Optional[Path] = typer.Option(
        None,
        "--destination",
        "-d",
        help="Path of the new config file to create. Uses the default config path if not specified.",
    ),
    overwrite: bool = typer.Option(
        False, "--overwrite", help="Overwrite destination config file if it exists."
    ),
    legacy_json: bool = typer.Option(
        False,
        "--legacy-json-format",
        help="Use legacy JSON format mode in the new config file.",
    ),
) -> None:
    """Migrate a legacy .conf config to a new .toml config.

    The new config file will be created in the default location if no destination is specified.
    The new config enables the new JSON format by default.
    """
    from zabbix_cli.config.constants import DEFAULT_CONFIG_FILE
    from zabbix_cli.config.model import Config

    if source:
        conf = Config.from_file(source)
    else:
        if not app.state.is_config_loaded:
            # this should never happen!
            exit_err(
                "Application was unable to load a config. Use [option]--source[/] to specify one."
            )
        conf = app.state.config

    if not conf.app.is_legacy:
        exit_err(
            "Unable to detect legacy config file. Use [option]--source[/] to specify one."
        )

    if not destination:
        destination = DEFAULT_CONFIG_FILE
    if not destination.suffix == ".toml":
        destination = destination.with_suffix(".toml")

    if destination.exists() and not overwrite:
        exit_err(
            f"File {destination} already exists. Use [option]--overwrite[/] to overwrite it."
        )

    # Set the legacy JSON format flag in the new file
    # By default, we move users over to the new format.
    conf.app.legacy_json_format = legacy_json

    conf.dump_to_file(destination)
    success(f"Config migrated to {destination}")


@app.command("open", rich_help_panel=HELP_PANEL)
def open_config_dir(
    ctx: typer.Context,
    directory_type: DirectoryType = typer.Argument(
        help="The type of directory to open.",
        case_sensitive=False,
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="LINUX: Try to open with [option]--command[/] even if no window manager is detected.",
    ),
    path: bool = typer.Option(
        False,
        "--path",
        help="Show path instead of opening directory.",
    ),
    open_command: Optional[str] = typer.Option(
        None,
        "--command",
        help="Specify command to use to use for opening.",
    ),
) -> None:
    """Open an app directory in the system's file manager.

    Use --force to attempt to open when no DISPLAY env var is set.
    """
    # Try to load the config, but don't fail if it's not available
    try:
        config = app.state.config
    except ZabbixCLIError:
        config = None

    directory = get_directory(directory_type, config)
    if path:
        print_path(directory)
    else:
        open_directory(directory, command=open_command, force=force)


@app.command("sample_config", rich_help_panel=HELP_PANEL)
def sample_config(ctx: typer.Context) -> None:
    """Print a sample configuration file."""
    # Load the entire history, then limit afterwards
    from zabbix_cli.config.model import Config

    conf = Config.sample_config()
    print_toml(conf.as_toml())


@app.command(
    "show_zabbixcli_config", rich_help_panel=HELP_PANEL, hidden=True, deprecated=True
)
@app.command("show_config", rich_help_panel=HELP_PANEL)
def show_config(
    ctx: typer.Context,
    secrets: SecretMode = typer.Option(
        SecretMode.MASK,
        "--secrets",
        help="Display mode for secrets.",
        case_sensitive=False,
    ),
) -> None:
    """Show the current application configuration."""
    config = app.state.config
    print_toml(config.as_toml(secrets=secrets))
    if config.config_path:
        info(f"Config file: {config.config_path.absolute()}")


@app.command("show_dirs", rich_help_panel=HELP_PANEL)
def show_directories(ctx: typer.Context) -> None:
    """Show the default directories used by the application."""
    from zabbix_cli.commands.results.cli import DirectoriesResult

    result = DirectoriesResult.from_directory_types(list(DirectoryType))
    render_result(result)


@app.command("show_history", rich_help_panel=HELP_PANEL)
def show_history(
    ctx: typer.Context,
    limit: int = OPTION_LIMIT,
    # TODO: Add --session option to limit to current session
    # In order to add that, we need to store the history len at the start of the session
) -> None:
    """Show the command history."""
    # Load the entire history, then limit afterwards
    from zabbix_cli.commands.results.cli import HistoryResult

    history = list(app.state.history.get_strings())
    history = history[-limit:]
    render_result(HistoryResult(commands=history))


@app.command("update_config", rich_help_panel=HELP_PANEL)
def update_config(
    ctx: typer.Context,
    config_file: Optional[Path] = typer.Option(
        None, "--config-file", "-c", help="Location of the config file to update."
    ),
    secrets: SecretMode = typer.Option(
        SecretMode.PLAIN,
        "--secrets",
        help="DEPRECATED: Visibility mode for secrets.",
        hidden=True,
        case_sensitive=False,
    ),
    force: bool = typer.Option(False, "--force", help="Skip confirmation prompt."),
) -> None:
    """Update the config file with the current application state.

    Adds missing fields and updates deprecated fields to their new values.
    """
    from zabbix_cli.output.prompts import bool_prompt

    config_file = config_file or app.state.config.config_path
    if not config_file:
        exit_err("No config file specified and no config loaded.")
    if not force:
        if not bool_prompt("Update config file?", default=False):
            exit_err("Update cancelled.")

    # When we dump the config, we automatically exclude the deprecated fields
    # while taking advantage of the validators that update the new fields.
    config = app.state.config
    config.dump_updated_config(config_file=config_file)


@app.command("update", rich_help_panel=HELP_PANEL, hidden=True)
def update_application(ctx: typer.Context) -> None:
    """Update the application to the latest version.

    Primarily intended for use with PyInstaller builds, but can also be
    used for updating other installations (except Homebrew).
    """
    from zabbix_cli.__about__ import __version__
    from zabbix_cli.update import update

    info = update()
    if info and info.version:
        success(f"Application updated from {__version__} to {info.version}")
    else:
        success("Application updated.")



"""Base model for all configuration models."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict
from pydantic import PrivateAttr
from pydantic import ValidationInfo
from pydantic import field_validator
from pydantic import model_validator
from typing_extensions import Self

from zabbix_cli.config.utils import check_deprecated_fields


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(validate_assignment=True, extra="ignore")

    _deprecation_checked: bool = PrivateAttr(default=False)
    """Has performed a deprecaction check for the fields on the model."""

    @field_validator("*")
    @classmethod
    def _conf_bool_validator_compat(cls, v: Any, info: ValidationInfo) -> Any:
        """Handles old config files that specified bools as ON/OFF."""
        if not isinstance(v, str):
            return v
        if v.upper() == "ON":
            return True
        if v.upper() == "OFF":
            return False
        return v

    # TODO: move this validator to Config, so we can provide exact config
    # field names, as well as optionally silencing the warnings with a flag.
    @model_validator(mode="after")
    def _check_deprecated_fields(self) -> Self:
        """Check for deprecated fields and log warnings."""
        if not self._deprecation_checked:
            check_deprecated_fields(self)
            self._deprecation_checked = True
        return self



from __future__ import annotations

from pathlib import Path
from typing import Optional

import pytest
import typer
from inline_snapshot import snapshot
from zabbix_cli.app.app import StatefulApp
from zabbix_cli.bulk import BulkCommand
from zabbix_cli.bulk import BulkRunner
from zabbix_cli.bulk import BulkRunnerMode
from zabbix_cli.bulk import CommandExecution
from zabbix_cli.bulk import CommandResult
from zabbix_cli.bulk import CommentLine
from zabbix_cli.bulk import EmptyLine
from zabbix_cli.exceptions import CommandFileError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import exit_ok
from zabbix_cli.output.console import info


@pytest.mark.parametrize(
    "line, expect",
    [
        pytest.param(
            "show_zabbixcli_config",
            BulkCommand(args=["show_zabbixcli_config"], line="show_zabbixcli_config"),
            id="simple",
        ),
        pytest.param(
            "create_user username name surname passwd role autologin autologout groups",
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "name",
                    "surname",
                    "passwd",
                    "role",
                    "autologin",
                    "autologout",
                    "groups",
                ],
                line="create_user username name surname passwd role autologin autologout groups",
            ),
            id="Legacy positional args",
        ),
        pytest.param(
            "create_user username --firstname name --lastname surname --passwd mypass --role 1  --autologin --autologout 86400 --groups '1,2'",
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "--firstname",
                    "name",
                    "--lastname",
                    "surname",
                    "--passwd",
                    "mypass",
                    "--role",
                    "1",
                    "--autologin",
                    "--autologout",
                    "86400",
                    "--groups",
                    "1,2",
                ],
                line="create_user username --firstname name --lastname surname --passwd mypass --role 1  --autologin --autologout 86400 --groups '1,2'",
            ),
            id="args and kwargs",
        ),
        pytest.param(
            "create_user username myname --passwd mypass surname",
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "myname",
                    "--passwd",
                    "mypass",
                    "surname",
                ],
                line="create_user username myname --passwd mypass surname",
            ),
            id="kwarg between args",
        ),
        pytest.param(
            "create_user myuser --firstname myname --passwd mypasswd --role 1 # comment here --option value",
            BulkCommand(
                args=[
                    "create_user",
                    "myuser",
                    "--firstname",
                    "myname",
                    "--passwd",
                    "mypasswd",
                    "--role",
                    "1",
                ],
                line="create_user myuser --firstname myname --passwd mypasswd --role 1 # comment here --option value",
            ),
            id="Trailing comment",
        ),
        pytest.param(
            "",
            BulkCommand(),
            id="fails (empty)",
            marks=pytest.mark.xfail(raises=EmptyLine, strict=True),
        ),
        pytest.param(
            "#",
            BulkCommand(),
            id="fails (comment symbol)",
            marks=pytest.mark.xfail(raises=CommentLine, strict=True),
        ),
        pytest.param(
            "# create_user myuser myname mypasswd --role 1",
            BulkCommand(),
            id="fails (commented out line)",
            marks=pytest.mark.xfail(raises=CommentLine, strict=True),
        ),
    ],
)
def test_bulk_command_from_line(line: str, expect: BulkCommand) -> None:
    assert BulkCommand.from_line(line) == expect


def test_load_command_file(tmp_path: Path, ctx: typer.Context) -> None:
    """Test loading a command file."""
    file = tmp_path / "commands.txt"
    file.write_text(
        """# comment
show_zabbixcli_config # next line will be blank

create_user username --firstname name --lastname surname mypass 1 1 86400 1,2
create_user username --firstname name --lastname surname --passwd mypass --role 1  --autologin --autologout 86400 --groups '1,2'
# comment explaining the next command
create_user username --firstname name --lastname surname --passwd mypass # trailing comment
# Command with flag
acknowledge_event 123,456,789 --message "foo message" --close
# Command with negative flag
show_templategroup mygroup --no-templates
# Command with optional boolean flags
show_host *.example.com --no-maintenance --monitored
# Command with enum option (human-readable string)
show_host *.example.com --active available
# Command with enum option (API values)
show_host *.example.com --active 0
show_host *.example.com --active 1
show_host *.example.com --active 2
# we will end with a blank line
"""
    )
    b = BulkRunner(ctx, file)
    commands = b.load_command_file()
    assert len(commands) == snapshot(11)
    assert commands == snapshot(
        [
            BulkCommand(
                args=["show_zabbixcli_config"],
                line="show_zabbixcli_config # next line will be blank",
                line_number=2,
            ),
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "--firstname",
                    "name",
                    "--lastname",
                    "surname",
                    "mypass",
                    "1",
                    "1",
                    "86400",
                    "1,2",
                ],
                line="create_user username --firstname name --lastname surname mypass 1 1 86400 1,2",
                line_number=4,
            ),
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "--firstname",
                    "name",
                    "--lastname",
                    "surname",
                    "--passwd",
                    "mypass",
                    "--role",
                    "1",
                    "--autologin",
                    "--autologout",
                    "86400",
                    "--groups",
                    "1,2",
                ],
                line="create_user username --firstname name --lastname surname --passwd mypass --role 1  --autologin --autologout 86400 --groups '1,2'",
                line_number=5,
            ),
            BulkCommand(
                args=[
                    "create_user",
                    "username",
                    "--firstname",
                    "name",
                    "--lastname",
                    "surname",
                    "--passwd",
                    "mypass",
                ],
                line="create_user username --firstname name --lastname surname --passwd mypass # trailing comment",
                line_number=7,
            ),
            BulkCommand(
                args=[
                    "acknowledge_event",
                    "123,456,789",
                    "--message",
                    "foo message",
                    "--close",
                ],
                line='acknowledge_event 123,456,789 --message "foo message" --close',
                line_number=9,
            ),
            BulkCommand(
                args=["show_templategroup", "mygroup", "--no-templates"],
                line="show_templategroup mygroup --no-templates",
                line_number=11,
            ),
            BulkCommand(
                args=["show_host", "*.example.com", "--no-maintenance", "--monitored"],
                line="show_host *.example.com --no-maintenance --monitored",
                line_number=13,
            ),
            BulkCommand(
                args=["show_host", "*.example.com", "--active", "available"],
                line="show_host *.example.com --active available",
                line_number=15,
            ),
            BulkCommand(
                args=["show_host", "*.example.com", "--active", "0"],
                line="show_host *.example.com --active 0",
                line_number=17,
            ),
            BulkCommand(
                args=["show_host", "*.example.com", "--active", "1"],
                line="show_host *.example.com --active 1",
                line_number=18,
            ),
            BulkCommand(
                args=["show_host", "*.example.com", "--active", "2"],
                line="show_host *.example.com --active 2",
                line_number=19,
            ),
        ]
    )


@pytest.mark.parametrize(
    "mode",
    [BulkRunnerMode.STRICT, BulkRunnerMode.CONTINUE],
)
def test_bulk_runner_mode_invalid_line_strict(
    tmp_path: Path, ctx: typer.Context, mode: BulkRunnerMode
) -> None:
    """Test loading a command file with invalid lines in strict/continue mode."""
    file = tmp_path / "commands.txt"
    file.write_text(
        """\
# comment
show_host "*.example.com # Missing closing quote
"""
    )
    b = BulkRunner(ctx, file, mode=mode)
    with pytest.raises(CommandFileError):
        b.load_command_file()


def test_bulk_runner_mode_invalid_line_skip(tmp_path: Path, ctx: typer.Context) -> None:
    """Test loading a command file with invalid lines in skip mode."""
    file = tmp_path / "commands.txt"
    file.write_text(
        """\
# comment
show_host *.example.com --active 0
show_host "*.example.com # Missing closing quote
create_host foo.example.com --hostgroup "Linux servers"
"""
    )
    b = BulkRunner(ctx, file, mode=BulkRunnerMode.SKIP)
    commands = b.load_command_file()
    assert len(commands) == 2  # show_host, create_host
    assert len(b.executions) == 0
    assert len(b.skipped) == 2  # comment, invalid command

    # First line is comment
    # Second line is invalid command
    result = b.skipped[1]
    assert result.command == snapshot(
        BulkCommand(
            line='show_host "*.example.com # Missing closing quote', line_number=3
        )
    )
    assert result.result == CommandResult.SKIPPED
    assert repr(result.error) == snapshot("ValueError('No closing quotation')")
    assert result.line_number == snapshot(3)


def test_load_command_file_not_found(tmp_path: Path, ctx: typer.Context) -> None:
    """Test loading a command file that does not exist."""
    file = tmp_path / "commands.txt"
    assert not file.exists()
    b = BulkRunner(ctx, file)
    with pytest.raises(CommandFileError):
        b.load_command_file()


@pytest.mark.parametrize("mode", [BulkRunnerMode.STRICT, BulkRunnerMode.CONTINUE])
def test_bulk_runner_exit_code_handling(
    tmp_path: Path, app: StatefulApp, ctx: typer.Context, mode: BulkRunnerMode
) -> None:
    """Test handling of exit codes."""
    file = tmp_path / "commands.txt"
    file.write_text(
        """\
# comment
no_exit
exits_ok
exits_error
"""
    )

    @app.command(name="exits_ok")
    def exits_ok() -> None:
        exit_ok("This command exits with code 0")

    @app.command(name="exits_error")
    def exits_error() -> None:
        exit_err("This command exits with code 1")

    @app.command(name="no_exit")
    def on_exit() -> None:
        info("We just print a message here")

    cmd = typer.main.get_command(app)
    ctx.command = cmd

    b = BulkRunner(ctx, file, mode)
    with pytest.raises(CommandFileError) as excinfo:
        b.run_bulk()

    assert len(b.executions) == 3
    assert b.executions[0].result == CommandResult.SUCCESS
    assert b.executions[1].result == CommandResult.SUCCESS
    assert b.executions[2].result == CommandResult.FAILURE

    # Differing error messages between strict and continue
    exc = excinfo.exconly()
    if mode == BulkRunnerMode.STRICT:
        assert (
            exc
            == "zabbix_cli.exceptions.CommandFileError: Command failed: [command]exits_error[/]: 1"
        )
    else:
        assert (
            exc
            == "zabbix_cli.exceptions.CommandFileError: 1 commands failed:\nLine 4: [command]exits_error[/] [i](1)[/]"
        )


@pytest.mark.parametrize(
    "mode", [BulkRunnerMode.STRICT, BulkRunnerMode.CONTINUE, BulkRunnerMode.SKIP]
)
def test_bulk_commands_complex(
    tmp_path: Path, app: StatefulApp, ctx: typer.Context, mode: BulkRunnerMode
) -> None:
    """Test bulk execution of commands with multiple options and arguments."""
    file = tmp_path / "commands.txt"
    file.write_text(
        """\
# comment
# Every type of option and argument
mixed_command "some arg" "optional arg" --opt value --reqopt 42 --flag --boolopt
mixed_command "some arg" "optional arg" --opt value --reqopt 42 --flag --no-boolopt
# Again with short options (no optional arg)
mixed_command "some arg" -O value -R 123 -F --boolopt
mixed_command "some arg" -O value -R 123 -F --no-boolopt
# Omit optional options
mixed_command "some arg" -O value -R 123
# Only required arguments/options
mixed_command "some arg" --reqopt 42
# Only arguments (with quotes)
only_args "arg1" "arg2" "arg3"
# Only arguments (no quotes)
only_args arg1 arg2 arg3
# Only options
only_options --opt1 value --opt2 value --opt3 42 --opt4 42
only_options -O "str" -S "Optional[str]" -I 42 -N 42
"""
    )

    @app.command(name="mixed_command")
    def mixed_command(
        ctx: typer.Context,
        reqarg: str = typer.Argument(),
        optarg: Optional[str] = typer.Argument(None),
        opt: Optional[str] = typer.Option(None, "--opt", "-O"),
        reqopt: int = typer.Option(
            ...,  # type: ignore
            "--reqopt",
            "-R",
        ),
        flag: bool = typer.Option(False, "--flag", "-F"),
        boolopt: Optional[bool] = typer.Option(
            False,
            # Not specifying anything here should generate the options
            # --boolopt / --no-boolopt
        ),
    ) -> None:
        """Command with every type of option and argument"""
        exit_ok("Running mixed_command")

    @app.command(name="only_args")
    def only_args(
        ctx: typer.Context,
        arg1: str = typer.Argument(),
        arg2: str = typer.Argument("default value"),
        arg3: Optional[str] = typer.Argument(None),
    ) -> None:
        exit_ok("Running only_args")

    @app.command(name="only_options")
    def only_options(
        ctx: typer.Context,
        opt1: str = typer.Option(..., "--opt1", "-O"),
        opt2: Optional[str] = typer.Option(None, "--opt2", "-S"),
        opt3: int = typer.Option(..., "--opt3", "-I"),
        opt4: Optional[int] = typer.Option(None, "--opt4", "-N"),
    ) -> None:
        exit_ok("Running only_options")

    cmd = typer.main.get_command(app)
    ctx.command = cmd

    b = BulkRunner(ctx, file, mode)
    b.run_bulk()

    assert len(b.executions) == snapshot(10)
    assert b.executions == snapshot(
        [
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "mixed_command",
                        "some arg",
                        "optional arg",
                        "--opt",
                        "value",
                        "--reqopt",
                        "42",
                        "--flag",
                        "--boolopt",
                    ],
                    line='mixed_command "some arg" "optional arg" --opt value --reqopt 42 --flag --boolopt',
                    line_number=3,
                ),
                result=CommandResult.SUCCESS,
                line_number=3,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "mixed_command",
                        "some arg",
                        "optional arg",
                        "--opt",
                        "value",
                        "--reqopt",
                        "42",
                        "--flag",
                        "--no-boolopt",
                    ],
                    line='mixed_command "some arg" "optional arg" --opt value --reqopt 42 --flag --no-boolopt',
                    line_number=4,
                ),
                result=CommandResult.SUCCESS,
                line_number=4,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "mixed_command",
                        "some arg",
                        "-O",
                        "value",
                        "-R",
                        "123",
                        "-F",
                        "--boolopt",
                    ],
                    line='mixed_command "some arg" -O value -R 123 -F --boolopt',
                    line_number=6,
                ),
                result=CommandResult.SUCCESS,
                line_number=6,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "mixed_command",
                        "some arg",
                        "-O",
                        "value",
                        "-R",
                        "123",
                        "-F",
                        "--no-boolopt",
                    ],
                    line='mixed_command "some arg" -O value -R 123 -F --no-boolopt',
                    line_number=7,
                ),
                result=CommandResult.SUCCESS,
                line_number=7,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=["mixed_command", "some arg", "-O", "value", "-R", "123"],
                    line='mixed_command "some arg" -O value -R 123',
                    line_number=9,
                ),
                result=CommandResult.SUCCESS,
                line_number=9,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=["mixed_command", "some arg", "--reqopt", "42"],
                    line='mixed_command "some arg" --reqopt 42',
                    line_number=11,
                ),
                result=CommandResult.SUCCESS,
                line_number=11,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=["only_args", "arg1", "arg2", "arg3"],
                    line='only_args "arg1" "arg2" "arg3"',
                    line_number=13,
                ),
                result=CommandResult.SUCCESS,
                line_number=13,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=["only_args", "arg1", "arg2", "arg3"],
                    line="only_args arg1 arg2 arg3",
                    line_number=15,
                ),
                result=CommandResult.SUCCESS,
                line_number=15,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "only_options",
                        "--opt1",
                        "value",
                        "--opt2",
                        "value",
                        "--opt3",
                        "42",
                        "--opt4",
                        "42",
                    ],
                    line="only_options --opt1 value --opt2 value --opt3 42 --opt4 42",
                    line_number=17,
                ),
                result=CommandResult.SUCCESS,
                line_number=17,
            ),
            CommandExecution(
                command=BulkCommand(
                    args=[
                        "only_options",
                        "-O",
                        "str",
                        "-S",
                        "Optional[str]",
                        "-I",
                        "42",
                        "-N",
                        "42",
                    ],
                    line='only_options -O "str" -S "Optional[str]" -I 42 -N 42',
                    line_number=18,
                ),
                result=CommandResult.SUCCESS,
                line_number=18,
            ),
        ]
    )
    assert len(b.skipped) == snapshot(8)
    assert b.skipped == snapshot(
        [
            CommandExecution(
                command=BulkCommand(line="# comment", line_number=1),
                result=CommandResult.SKIPPED,
                line_number=1,
            ),
            CommandExecution(
                command=BulkCommand(
                    line="# Every type of option and argument", line_number=2
                ),
                result=CommandResult.SKIPPED,
                line_number=2,
            ),
            CommandExecution(
                command=BulkCommand(
                    line="# Again with short options (no optional arg)", line_number=5
                ),
                result=CommandResult.SKIPPED,
                line_number=5,
            ),
            CommandExecution(
                command=BulkCommand(line="# Omit optional options", line_number=8),
                result=CommandResult.SKIPPED,
                line_number=8,
            ),
            CommandExecution(
                command=BulkCommand(
                    line="# Only required arguments/options", line_number=10
                ),
                result=CommandResult.SKIPPED,
                line_number=10,
            ),
            CommandExecution(
                command=BulkCommand(
                    line="# Only arguments (with quotes)", line_number=12
                ),
                result=CommandResult.SKIPPED,
                line_number=12,
            ),
            CommandExecution(
                command=BulkCommand(
                    line="# Only arguments (no quotes)", line_number=14
                ),
                result=CommandResult.SKIPPED,
                line_number=14,
            ),
            CommandExecution(
                command=BulkCommand(line="# Only options", line_number=16),
                result=CommandResult.SKIPPED,
                line_number=16,
            ),
        ]
    )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Optional

from pydantic import computed_field
from typing_extensions import TypedDict

from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.enums import HostgroupFlag
from zabbix_cli.pyzabbix.enums import HostgroupType
from zabbix_cli.pyzabbix.types import Host
from zabbix_cli.pyzabbix.types import HostGroup
from zabbix_cli.pyzabbix.types import HostList

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType


class AddHostsToHostGroup(TableRenderable):
    """Result type for `add_host_to_hostgroup` and `remove_host_from_hostgroup` commands."""

    hostgroup: str
    hosts: list[str]

    @classmethod
    def from_result(
        cls,
        hosts: list[Host],
        hostgroup: HostGroup,
    ) -> AddHostsToHostGroup:
        to_add: set[str] = set()  # names of templates to link
        for host in hosts:
            for hg_host in hostgroup.hosts:
                if host.host == hg_host.host:
                    break
            else:
                to_add.add(host.host)
        return cls(
            hostgroup=hostgroup.name,
            hosts=sorted(to_add),
        )


class RemoveHostsFromHostGroup(TableRenderable):
    """Result type for `remove_host_from_hostgroup`."""

    hostgroup: str
    hosts: list[str]

    @classmethod
    def from_result(
        cls,
        hosts: list[Host],
        hostgroup: HostGroup,
    ) -> RemoveHostsFromHostGroup:
        to_remove: set[str] = set()  # names of templates to link
        for host in hosts:
            for hg_host in hostgroup.hosts:
                if host.host == hg_host.host:
                    to_remove.add(host.host)
                    break
        return cls(
            hostgroup=hostgroup.name,
            hosts=sorted(to_remove),
        )


class ExtendHostgroupResult(TableRenderable):
    """Result type for `extend_hostgroup` command."""

    source: str
    destination: list[str]
    hosts: list[str]

    @classmethod
    def from_result(
        cls, source: HostGroup, destination: list[HostGroup]
    ) -> ExtendHostgroupResult:
        return cls(
            source=source.name,
            destination=[dst.name for dst in destination],
            hosts=[host.host for host in source.hosts],
        )


class MoveHostsResult(TableRenderable):
    """Result type for `move_hosts` command."""

    source: str
    destination: str
    hosts: list[str]

    @classmethod
    def from_result(cls, source: HostGroup, destination: HostGroup) -> MoveHostsResult:
        return cls(
            source=source.name,
            destination=destination.name,
            hosts=[host.host for host in source.hosts],
        )


class HostGroupDeleteResult(TableRenderable):
    groups: list[str]


class HostGroupHost(TypedDict):
    hostid: str
    host: str


class HostGroupResult(TableRenderable):
    """Result type for hostgroup."""

    groupid: str
    name: str
    hosts: HostList = []
    flags: int
    internal: Optional[int] = None  # <6.2

    @classmethod
    def from_hostgroup(cls, hostgroup: HostGroup) -> HostGroupResult:
        return cls(
            groupid=hostgroup.groupid,
            name=hostgroup.name,
            flags=hostgroup.flags,
            internal=hostgroup.internal,  # <6.2
            hosts=hostgroup.hosts,
        )

    # LEGACY
    # Mimicks old behavior by also writing the string representation of the
    # flags and internal fields to the serialized output.
    @computed_field
    @property
    def flags_str(self) -> str:
        return HostgroupFlag.string_from_value(self.flags, with_code=False)

    # VERSION: 6.0
    # Internal groups are not a thing in Zabbix >=6.2
    @computed_field
    @property
    def type(self) -> str:
        return HostgroupType.string_from_value(self.internal, with_code=True)

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Name", "Flag", "Hosts"]
        rows: RowsType = [
            [
                self.groupid,
                self.name,
                self.flags_str,
                ", ".join(host.host for host in self.hosts),
            ]
        ]
        # VERSION: 6.0
        if self.zabbix_version.release < (6, 2):
            cols.insert(3, "Type")
            t = HostgroupType.string_from_value(self.internal, with_code=False)
            rows[0].insert(3, t)  # without code in table
        return cols, rows


class HostGroupPermissions(TableRenderable):
    """Result type for hostgroup permissions."""

    groupid: str
    name: str
    permissions: list[str]

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["GroupID", "Name", "Permissions"]
        rows: RowsType = [[self.groupid, self.name, "\n".join(self.permissions)]]
        return cols, rows
